from collections import namedtuple

def get_options():

    _input_norm_args = {'abs_max_scaler': {'cls': 'MaxAbsScaler',
                                           'params': {
                                               'copy': True}},
                        'min_max_scaler': {'cls': 'MinMaxScaler',
                                           'params': {
                                               'feature_range': (-1, 1),
                                               'copy': True,
                                               'clip': False}},
                        'l1_norm': {'cls': 'Normalizer',
                                    'params': {
                                        'norm': 'l1',
                                        'copy': True}},
                        'l2_norm': {'cls': 'Normalizer',
                                    'params': {
                                        'norm': 'l2',
                                        'copy': True}},
                        'power_trans': {'cls': 'PowerTransformer',
                                        'params': {
                                            'method': 'yeo-johnson',
                                            'standardize': True,
                                            'copy': True}},
                        'quant_trans': {'cls': 'QuantileTransformer',
                                        'params': {
                                            'n_quantiles': 1000,
                                            'output_distribution': 'uniform',
                                            'ignore_implicit_zeros': False,
                                            'subsample': 100000,
                                            'random_state': None,
                                            'copy': True}},
                        'robust_scaler': {'cls': 'RobustScaler',
                                          'params': {
                                              'with_centering': True,
                                              'with_scaling': True,
                                              'quantile_range': (25.0, 75.0),
                                              'copy': True,
                                              'unit_variance': False}},
                        'standard_scaler': {'cls': 'StandardScaler',
                                            'params': {
                                                'copy': True,
                                                'with_mean': True,
                                                'with_std': True}}}

    _model_reg_args = {'drop': {'type': 'drop', 'args': {}},
                       'spa_drop': {'type': 'spa-drop', 'args': {}},
                       'gauss_drop': {'type': 'gauss-drop', 'args': {}},
                       'alpha_drop': {'type': 'alpha-drop', 'args': {}},
                       'gauss_noise': {'type': 'gauss-noise', 'args': {}},
                       'act_reg': {'type': 'act-reg', 'args': {}}}

    _model_norm_args = {'layer': {'type': 'layer', 'args': {}},
                        'batch': {'type': 'batch', 'args': {}}}

    InputNormArgs = namedtuple('InputNormArgs', [_ for _ in _input_norm_args])
    ModelRegArgs = namedtuple('ModelRegArgs', [_ for _ in _model_reg_args])
    ModelNormArgs = namedtuple('ModelNormArgs', [_ for _ in _model_norm_args])

    Options = namedtuple('Options', ['input_norm_args','model_reg_args','model_norm_args'])

    return Options(input_norm_args=InputNormArgs(**_input_norm_args),
                   model_reg_args=ModelRegArgs(**_model_reg_args),
                   model_norm_args=ModelNormArgs(**_model_norm_args))


import tensorflow as tf

from dataclasses import dataclass, field, replace as dc_replace
from typing import List, Dict, Any
from datetime import datetime as dt

def setup_runs(network, branches, run_config, run_fn, rebuild_on):

    if not run_fn:
        run_fn = 'data-models'

    # -- init run -- #
    run = Runs(network, run_config, run_fn, rebuild_on)

    run.run_idx = network.n_runs

    # -- loop network branches -- #
    for branch_idx, branch in enumerate(branches):

        run.add_branch(run, branch, branch_idx)

        # -- loop network chains -- #
        for chain_idx, chain in enumerate(branch.chains):

            run.branches[branch_idx].add_chain(chain, chain_idx)

            branch_run_config = branch.configs.by_chain_idx('run', chain_idx)

            _config = {'run_fn': run_fn,
                       'run_type': branch_run_config['run_type'],
                       'data_type': branch_run_config['data_type'],
                       'val_on': branch_run_config['val_on'],
                       'metrics_on': branch_run_config['metrics_on'],
                       'pre_build': branch_run_config['pre_build'],
                       'load_cp': branch_run_config['load_cp'],
                       'save_cp': branch_run_config['save_cp'],
                       'msgs_on': branch_run_config['msgs_on'],
                       'async_on': branch_run_config['async_on'],
                       'n_metrics_fns': len(branch.configs.by_chain_idx('metrics', chain_idx))}

            if len(run_config) > chain_idx:

                new_run_config = run_config[chain_idx]

                new_run_keys = new_run_config.keys()

                for _key in new_run_keys:

                    if _key != 'n_epochs':
                        _config[_key] = new_run_config[_key]

                    # -- override n_epochs directly -- #
                    if _key == 'n_epochs':
                        branch.n_epochs = new_run_config['n_epochs']

                    # -- override branch msgs_on directly -- #
                    if _key == 'msgs_on' and _config[_key]:
                        branch.msgs_on = True

            run.branches[branch_idx].chains[chain_idx].add_config(_config)

            run.branches[branch_idx].chains[chain_idx].add_live()

            for model_idx, gauge in enumerate(chain.models):

                run.branches[branch_idx].chains[chain_idx].add_model(gauge, model_idx)

                run.branches[branch_idx].chains[chain_idx].models[model_idx].add_live()
                run.branches[branch_idx].chains[chain_idx].models[model_idx].add_steps()
                run.branches[branch_idx].chains[chain_idx].models[model_idx].build_model()

    network.runs.append(run)
    network.n_runs += 1

    return run

@dataclass
class Runs:

    network: Any
    run_config: List
    run_fn: str
    rebuild_on: bool

    id: int = int(dt.now().timestamp())

    branches: List = field(default_factory=lambda:[])

    def add_branch(self, run, branch, branch_idx):

        run_branch = Branches(run, branch, branch_idx)

        run_branch.src.run = run_branch

        self.branches.insert(branch_idx, run_branch)

    def __repr__(self):
        return f'{self.__class__} {hex(id(self))}'

@dataclass
class Branches:

    run: Runs
    src: Any
    branch_idx: int
    chains: List = field(default_factory=lambda:[])
    batch_data: List = field(default_factory=lambda:[])
    epoch_msg: Any = None

    def add_chain(self, chain, chain_idx):

        self.chains.insert(chain_idx, Chains(self, chain, chain_idx, chain.n_outputs))

    def __repr__(self):
        return f'{self.__class__} {hex(id(self))}'

@dataclass
class Chains:

    branch: Any
    src: Any
    chain_idx: int
    n_outputs: int
    n_models: int = 0
    models: List = field(default_factory=lambda:[])

    def add_config(self, config):

        self.config = self.Configs(chain=self, **config)

    def add_live(self):

        self.live = self.Live(self)

    def add_model(self, gauge, model_idx):

        pre_step_idx = 0

        # --- if previous runs and not rebuild on --- #
        if self.branch.run.network.runs and not self.branch.run.rebuild_on:

            # --- get/set pre_step_idx --- #

            pre_run = self.branch.run.network.runs[-1]
            pre_step_idx = \
                pre_run.branches[self.branch.branch_idx].chains[self.chain_idx].models[model_idx].live.step_idx

        # -- insert models into chain class models -- #
        self.models.insert(model_idx, self.Models(self, gauge, gauge.src, model_idx, pre_step_idx))

        self.n_models += 1

    def __repr__(self):
        return f'{self.__class__} {hex(id(self))}'

    @dataclass
    class Configs:

        chain: Any
        run_fn: str
        run_type: str
        data_type: str
        val_on: bool
        metrics_on: bool
        pre_build: bool
        load_cp: bool
        save_cp: bool
        msgs_on: bool
        async_on: bool
        n_metrics_fns: int

        def __repr__(self):
            return f'{self.__class__} {hex(id(self))}'

    @dataclass
    class Live:

        chain: Any

        run_type: str = None
        data_type: str = None
        is_val: str = None

        epoch_idx: int = None
        batch_idx: int = None
        tree_idx: int = 0

        def __repr__(self):
            return f'{self.__class__} {hex(id(self))}'

        @classmethod
        def set_cls_attr(cls, xxx):
            if xxx:
                cls.cls_attr = 'this_value'
            else:
                cls.cls_attr = 'that_value'

    @dataclass
    class Models:

        chain: Any
        gauge: Any
        src: Any
        model_idx: int

        pre_step_idx: int

        steps_log: List = field(default_factory=lambda:[])

        def add_live(self):
            self.live = self.Live(self)

        def add_steps(self):

            self.steps = self.Steps(self)
            self.val_steps = self.Steps(self)

        def build_model(self):

            if not self.gauge.is_built:
                self.gauge.build_gauge(self.chain.branch.run, self.chain, self)

            self.src = self.gauge.src

            # self.gauge.setup_run(self.chain.branch.run, self.chain, self)

            # -- if first run of network or rebuild is on -- #
            if not self.chain.branch.run.network.runs or self.chain.branch.run.rebuild_on:

                # --- setup cp and rebuild --- #

                if self.gauge.strat_on:

                    with self.gauge.strat.scope():

                        if not self.gauge.is_compiled:

                            self.gauge.compile_gauge(self.chain.live.tree_idx)

                        if not self.chain.branch.run.network.runs:

                            if self.chain.config.load_cp or self.chain.config.save_cp:

                                self.gauge.setup_cp(self.live.step_idx, self.chain.config.load_cp)

                        if self.chain.config.pre_build:

                            self.gauge.pre_build_model()

                if not self.gauge.strat_on:

                    if not self.gauge.is_compiled:
                        self.gauge.compile_gauge(self.chain.live.tree_idx)

                    if not self.chain.branch.run.network.runs:

                        if self.chain.config.load_cp or self.chain.config.save_cp:
                            self.gauge.setup_cp(self.live.step_idx, self.chain.config.load_cp)

                    if self.chain.config.pre_build:
                        self.gauge.pre_build_model()

        def __repr__(self):
            return f'{self.__class__} {hex(id(self))}'

        @dataclass
        class Live:

            model: Any

            run_type: str = None
            data_type: str = None
            is_val: bool = None
            is_training: bool = None

            epoch_idx: int = None
            batch_idx: int = None
            model_idx: int = None

            batch_data: List = field(default_factory=lambda:[])
            val_batch_data: List = field(default_factory=lambda: [])

            def __post_init__(self):

                self.step_idx = tf.Variable(self.model.pre_step_idx,
                                            name=self.model.gauge.name + '_step_idx',
                                            dtype=tf.int32,
                                            trainable=False)

            def __repr__(self):
                return f'{self.__class__} {hex(id(self))}'

        @dataclass
        class Steps:

            model: Any

            specs: List = field(default_factory=lambda: [[]])

            step_loss: List = field(default_factory=lambda: [[]])
            model_loss: List = field(default_factory=lambda: [[]])
            full_loss: List = field(default_factory=lambda: [[]])

            x_tracking: List = field(default_factory=lambda: [[]])

            y_true: List = field(default_factory=lambda: [[]])
            y_hat: List = field(default_factory=lambda: [[]])
            y_tracking: List = field(default_factory=lambda: [[]])

            learning_rates: List = field(default_factory=lambda: [[]])

            metrics: List = field(default_factory=lambda: [[]])

            grads: List = field(default_factory=lambda: [[]])

            features: List = field(default_factory=lambda: [[]])

            preds_dist: List = field(default_factory=lambda: [[]])
            preds_samples: List = field(default_factory=lambda: [[]])

            def __repr__(self):
                return f'{self.__class__} {hex(id(self))}'


import tensorflow as tf
import numpy as np

class AdamDynamic(tf.keras.optimizers.Adam):

    def __init__(self, gauge, tree, config, n_epochs, **kwargs):

        self.gauge = gauge
        self.tree = tree
        self.config = config
        self.n_epochs = n_epochs
        
        self.dtype_float = self.gauge.dtype

        self.logs = []

        lr_st = tf.Variable(self.config['lr_st'],
                                 name='lr_st',
                                 dtype=self.dtype_float,
                                 trainable=False)

        lr_min = tf.Variable(self.config['lr_min'],
                                  name='lr_min',
                                  dtype=self.dtype_float,
                                  trainable=False)

        decay_rate = tf.Variable(self.config['decay_rate'],
                                 name='decay_rate',
                                 dtype=self.dtype_float,
                                 trainable=False)

        batch_size = tf.Variable(self.tree.data.store['train']['config']['batch_size'],
                                 name='batch_size',
                                 dtype=self.dtype_float,
                                 trainable=False)

        n_epochs = tf.Variable(self.n_epochs,
                               name='n_epochs',
                               dtype=self.dtype_float,
                               trainable=False)

        n_samples = tf.Variable(self.tree.data.store['train']['config']['n_samples'],
                                name='n_samples',
                                dtype=self.dtype_float,
                                trainable=False)

        static_st = tf.Variable(self.config['static_epochs'][0],
                                  name='static_epochs_st',
                                  dtype=self.dtype_float,
                                  trainable=False)

        static_ed = tf.Variable(self.config['static_epochs'][1],
                                    name='static_epochs_ed',
                                    dtype=self.dtype_float,
                                    trainable=False)

        steps_pe = tf.Variable((n_samples / batch_size))

        n_steps = tf.Variable(steps_pe * n_epochs,
                              name='n_steps',
                              dtype=self.dtype_float,
                              trainable=False)

        self.sch_config = {'lr_st': lr_st,
                           'lr_min': lr_min,
                           'decay_rate': decay_rate,
                           'static_st': static_st,
                           'static_ed': static_ed,
                           'batch_size': batch_size,
                           'n_epochs': n_epochs,
                           'n_samples': n_samples,
                           'n_steps': n_steps,
                           'steps_pe': steps_pe}

        self.lr_sch = PhotonSch(self)

        super().__init__(learning_rate=self.lr_sch)

class PhotonSch(tf.keras.optimizers.schedules.LearningRateSchedule):

    def __init__(self, opt):

        self.opt = opt
        self.config = self.opt.sch_config

        static_decay = ((self.config['static_st'] * self.config['steps_pe']) -
                        (self.config['static_ed'] * self.config['steps_pe']))

        self.cur_lr = tf.Variable(self.config['lr_st'],
                                  name='cur_lr',
                                  dtype=self.opt.dtype_float,
                                  trainable=False)

        self.cur_step = tf.Variable(0,
                                    name='cur_step',
                                    dtype=self.opt.dtype_float,
                                    trainable=False)

        self.steps_left = tf.Variable(0,
                                      name='steps_left',
                                      dtype=self.opt.dtype_float,
                                      trainable=False)

        self.decay_rng = tf.Variable(self.config['lr_st'] - self.config['lr_min'],
                                     name='decay_rng',
                                     dtype=self.opt.dtype_float,
                                     trainable=False)

        self.decay_steps = tf.Variable((self.config['n_steps'] - static_decay + 1),
                                       name='decay_steps',
                                       dtype=self.opt.dtype_float,
                                       trainable=False)

        self.decay_ps = tf.Variable(self.decay_rng / self.decay_steps,
                                    name='decay_ps',
                                    dtype=self.opt.dtype_float,
                                    trainable=False)

        self.cur_lr.assign(self.config['lr_st'])

    def __call__(self, step):

        self.cur_step.assign_add(1)
        self.steps_left.assign_sub(1)

        _reduc = tf.maximum(self.decay_ps * (step - self.config['static_st'] + 1) * self.config['decay_rate'],0)

        self.cur_lr.assign(tf.maximum((self.config['lr_st'] - _reduc), self.config['lr_min']))

        return self.cur_lr


from typing import List, Tuple, Dict, Optional, Any
from photon import Gauge
from photon.utils import np_exp

import numpy as np
import tensorflow as tf

from tensorflow import math as tfm
from tensorflow.keras import layers as tf_layers
from tensorflow.keras.layers import Layer as tf_Layer
from sklearn import preprocessing

from tensorflow.keras import activations, initializers, regularizers, constraints

def get_shapes(inputs):

    input_shp = []
    input_dims = 0

    if isinstance(inputs, list):

        for i in inputs:
            input_shp.append(i.get_shape())

        input_dims = len(input_shp[0])

    else:

        input_shp = inputs.get_shape()
        input_dims = len(input_shp)

    return input_shp

def get_act(act_fn):

    if act_fn == 'tanh':
        return tf.nn.tanh

    if act_fn == 'relu':
        return tf.nn.relu

    if act_fn == 'softmax':
        return tf.nn.softmax

    if act_fn == 'sigmoid':
        return tf.nn.sigmoid

    if act_fn == 'elu':
        return tf.nn.elu

def del_key(data, key):
    if key in data:
        del data[key]
    return data

class Layers(tf_Layer):

    def __init__(self,
                 gauge: Gauge,
                 layer_nm: str,
                 no_subs: bool = False,
                 no_log: bool = False,
                 is_child: bool = False,
                 reg_args: Optional[Tuple] = None,
                 norm_args: Optional[Tuple] = None,
                 **kwargs):

        self.gauge = gauge
        self.layer_nm = layer_nm
        self.no_subs = no_subs
        self.no_log = no_log

        self.is_child = is_child

        self.reg_args = reg_args
        self.norm_args = norm_args
        self.reg_vals = self.gauge.model_args['reg_vals']

        self.layers_logs_on = 0

        self.gauge.chain.idx_gen.append(self)
        self.layer_idx = len(self.gauge.chain.idx_gen) - 1

        # -- save layers to gauge -- #

        self.gauge.layers[self.layer_nm] = self

        if not self.is_child:
            self.gauge.parent_layers[self.layer_nm] = self

        if self.is_child:
            self.gauge.child_layers[self.layer_nm] = self

        # -- reg/norm -- #

        self.reg_on = False
        self.norm_on = False

        # -- add reg  -- #
        if self.reg_args is not None and not self.no_subs:

            self.reg_on = True
            self.reg_layer_nm = self.layer_nm + '_reg'

            self.gauge.src.reg_layers[self.reg_layer_nm] = \
                Reg(self.gauge, layer_nm=self.layer_nm + '_reg', reg_type=self.reg_args['type'], is_child=True, **self.reg_args['args'])

        # -- add norm  -- #
        if self.norm_args is not None and not self.no_subs:

            self.norm_on = True
            self.norm_layer_nm = self.gauge.name.lower() + '_' + self.layer_nm + '_norm'

            self.gauge.src.norm_layers[self.norm_layer_nm] = \
                Norm(self.gauge, layer_nm=self.norm_layer_nm, norm_type=self.norm_args['type'], is_child=True, **self.norm_args['args'])

        self.logs_on = False

        if 'logs_on' in kwargs:
            self.logs_on = kwargs['logs_on']

        self.logs = []

        super().__init__(name=self.layer_nm)

    def __call__(self, inputs, training=None, **kwargs):

        if training is None:

            if not self.gauge.is_live:
                training = False

            if self.gauge.is_live:
                training = self.gauge.run_model.live.is_training

        if not self.built:

            self.input_shp = get_shapes(inputs)

            # -- call build -- #
            self.build(self.input_shp)
            self.built = True

        # -- call layer -- #
        z_output = self.call(inputs, training=training)

        if self.layers_logs_on:
            self.log_call(self.layer_nm, inputs, z_output)

        # -- call reg -- #
        if self.reg_on:

            if not self.gauge.src.reg_layers[self.reg_layer_nm].is_built:
                self.gauge.src.reg_layers[self.reg_layer_nm].build(z_output.shape)
                self.gauge.src.reg_layers[self.reg_layer_nm].is_built = True

            z_output = self.gauge.src.reg_layers[self.reg_layer_nm].call(z_output, training=training)

        # -- call norm -- #
        if self.norm_on:

            if not self.gauge.src.norm_layers[self.norm_layer_nm].is_built:
                self.gauge.src.norm_layers[self.norm_layer_nm].build(z_output.shape)
                self.gauge.src.norm_layers[self.norm_layer_nm].is_built = True

            z_output = self.gauge.src.norm_layers[self.norm_layer_nm].call(z_output, training=training)

        return z_output

    def log_call(self, layer_nm, inputs, z_output):

        if self.gauge.cur_run.data_type == 'train' and not self.no_log:

            chain_idx = self.gauge.cur_run.chain_idx
            epoch_idx =  self.gauge.cur_run.epoch_idx
            batch_idx =  self.gauge.cur_run.batch_idx

            log_in = inputs
            log_out = z_output

            log_in_sh = []
            log_out_sh = []

            if isinstance(log_in, list):

                for _in_log in log_in:
                    log_in_sh.append(_in_log.shape.as_list())

            else:

                log_in_sh = log_in.shape.as_list()

                if self.layers_logs_on == 1:
                    if log_in.__class__.__name__ == 'EagerTensor':
                        log_in = log_in.numpy()
                    else:
                        log_in = log_in

            if isinstance(log_out, list):

                for _out_log in log_out:
                    log_out_sh.append(_out_log.shape.as_list())
            else:
                log_out_sh = log_out.shape.as_list()

                if self.layers_logs_on == 1:

                    if log_out.__class__.__name__ == 'EagerTensor':
                        log_out = log_out.numpy()
                    else:
                        log_out = log_out

            if self.layers_logs_on == 1:

                log_data = {
                    'step': [chain_idx, epoch_idx, batch_idx],
                    'layer': self,
                    'layer_idx': self.layer_idx,
                    'layer_name': layer_nm,
                    'in_shape': log_in_sh,
                    'out_shape': log_out_sh,
                    'input': log_in,
                    'output': log_out,
                    'loss': [_loss.numpy() for _loss in self.losses]
                }

            if self.layers_logs_on == 2:

                log_data = {
                    'layer': self,
                    'layer_idx': self.layer_idx,
                    'layer_name': layer_nm,
                    'in_shape': log_in_sh,
                    'out_shape': log_out_sh
                }

            if len(self.gauge.logs.layers) <= epoch_idx:
                self.gauge.logs.layers.append([])

            if len(self.gauge.logs.layers[epoch_idx]) <= batch_idx:
                self.gauge.logs.layers[epoch_idx].append([])

            # -- log_data to save output -- #
            self.gauge.logs.layers[epoch_idx][batch_idx].append(log_data)

    def save_layer_log(self, _log):

        epoch_idx = self.gauge.cur_run.epoch_idx.numpy()
        batch_idx = self.gauge.cur_run.batch_idx.numpy()

        if len(self.logs) <= epoch_idx:
            self.logs.insert(epoch_idx, [])

        if len(self.logs[epoch_idx]) <= batch_idx:
            self.logs[epoch_idx].insert(batch_idx, {})

        self.logs[epoch_idx][batch_idx] = _log

class Reg(Layers):

    def __init__(self, gauge, layer_nm, reg_type, **kwargs):

        super().__init__(gauge, layer_nm, no_subs=True, **kwargs)

        self.reg_type = reg_type.lower()
        self.is_built = False

    def build(self, input_shp, **kwargs):

        if self.reg_type == 'drop':
            self.k_layer = tf_layers.Dropout(rate=self.reg_vals[0], name=self.layer_nm, **kwargs)

        if self.reg_type == 'spa-drop':
            self.k_layer = tf_layers.SpatialDropout1D(rate=self.reg_vals[0], name=self.layer_nm, **kwargs)

        if self.reg_type == 'gauss-drop':
            self.k_layer = tf_layers.GaussianDropout(rate=self.reg_vals[0], name=self.layer_nm, **kwargs)

        if self.reg_type == 'alpha-drop':
            self.k_layer = tf_layers.AlphaDropout(rate=self.reg_vals[0], name=self.layer_nm, **kwargs)

        if self.reg_type == 'gauss-noise':
            self.k_layer = tf_layers.GaussianNoise(stddev=self.reg_vals[0], name=self.layer_nm, **kwargs)

        if self.reg_type == 'act-reg':
            self.k_layer = tf_layers.ActivityRegularization(l1=self.reg_vals[0], l2=self.reg_vals[1], name=self.layer_nm, **kwargs)

        return

    def call(self, inputs, training=None, **kwargs):

        z_output = self.k_layer(inputs=inputs, training=training)

        return z_output

class Norm(Layers):

    def __init__(self, gauge, layer_nm, norm_type, **kwargs):

        super().__init__(gauge, layer_nm=layer_nm, **kwargs)

        self.gauge = gauge
        self.layer_nm = layer_nm
        self.norm_type = norm_type.lower()
        self.is_built = False

    def build(self, input_shp, **kwargs):

        self.input_shp = input_shp
        self.input_dims = len(input_shp)

        if self.norm_type == 'layer':
            self.k_layer = tf_layers.LayerNormalization(name=self.layer_nm)

        if self.norm_type == 'batch':
            self.k_layer = tf_layers.BatchNormalization(name=self.layer_nm)

        return

    def call(self, inputs, training=None, **kwargs):

        # print('calling norm layer: ', self.layer_nm, training, '\n')

        z_output = self.k_layer(inputs=inputs, training=training)

        return z_output

class Base(Layers):

    def __init__(self, gauge, layer_nm, layer, **kwargs):

        super().__init__(gauge, layer_nm, **kwargs)

        self.k_layer = layer

    def build(self, input_shp, **kwargs):

        self.input_shp = input_shp

        return

    def call(self, inputs, training=None, **kwargs):

        return self.k_layer(inputs=inputs, training=training)

class DNN(Layers):

    def __init__(self, gauge, layer_nm, layer_args, **kwargs):

        super().__init__(gauge, layer_nm, **kwargs)

        self.layer_args = layer_args

    def build(self, input_shp):

        self.input_shp = input_shp

        self.k_layer = tf_layers.Dense(name=self.layer_nm, **self.layer_args)

        return

    def call(self, inputs, training=None, **kwargs):
        return self.k_layer(inputs=inputs, training=training)

class CNN(Layers):

    def __init__(self, gauge, layer_nm, layer_args, **kwargs):

        super().__init__(gauge, layer_nm, **kwargs)

        self.layer_args = layer_args

        self.filters = 1
        self.kernel_size = 1

        if 'filters' in kwargs:
            self.filters = kwargs['filters']
            self.layer_args = del_key(self.layer_args, 'filters')
        else:
            self.filters = self.layer_args['filters']

        if 'kernel_size' in kwargs:
            self.kernel_size = kwargs['kernel_size']
            self.layer_args = del_key(self.layer_args, 'kernel_size')
        else:
            self.kernel_size = self.layer_args['kernel_size']

        self.cnn_args = self.set_args(**self.layer_args)

    def build(self, input_shp):

        self.input_shp = input_shp
        self.k_layer = tf_layers.Conv1D(name=self.layer_nm, **self.cnn_args)

        return

    def call(self, inputs, training=None, **kwargs):

        return self.k_layer(inputs=inputs, training=training)

    def set_args(self, **kwargs):

        # -- default args -- #

        filters = self.filters
        kernel_size = self.kernel_size
        strides = 1
        padding = 'same'
        dilation_rate = 1,
        activation = 'tanh'
        use_bias = True
        kernel_initializer = 'glorot_uniform'
        bias_initializer = 'zeros'
        kernel_regularizer = None
        bias_regularizer = None
        activity_regularizer = None
        kernel_constraint = None
        bias_constraint = None
        data_format = 'channels_last'

        # -- args -- #

        if 'filters' in kwargs:
            filters = kwargs['filters']

        if 'kernel_size' in kwargs:
            kernel_size = kwargs['kernel_size']

        if 'strides' in kwargs:
            strides = kwargs['strides']

        if 'padding' in kwargs:
            padding = kwargs['padding']

        if 'dilation_rate' in kwargs:
            dilation_rate = kwargs['dilation_rate']

        if 'activation' in kwargs:
            activation = kwargs['activation']

        if 'use_bias' in kwargs:
            use_bias = kwargs['use_bias']

        if 'kernel_initializer' in kwargs:
            kernel_initializer = kwargs['kernel_initializer']

        if 'bias_initializer' in kwargs:
            bias_initializer = kwargs['bias_initializer']

        if 'kernel_regularizer' in kwargs:
            kernel_regularizer = kwargs['kernel_regularizer']

        if 'bias_regularizer' in kwargs:
            bias_regularizer = kwargs['bias_regularizer']

        if 'activity_regularizer' in kwargs:
            activity_regularizer = kwargs['activity_regularizer']

        if 'kernel_constraint' in kwargs:
            kernel_constraint = kwargs['kernel_constraint']

        if 'bias_constraint' in kwargs:
            bias_constraint = kwargs['bias_constraint']

        if 'data_format' in kwargs:
            data_format = kwargs['data_format']

        _args = {
            'filters': filters,
            'kernel_size': kernel_size,
            'strides': strides,
            'padding': padding,
            'dilation_rate': dilation_rate,
            'activation': activation,
            'use_bias': use_bias,
            'kernel_initializer': kernel_initializer,
            'bias_initializer': bias_initializer,
            'kernel_regularizer': kernel_regularizer,
            'bias_regularizer': bias_regularizer,
            'activity_regularizer': activity_regularizer,
            'kernel_constraint': kernel_constraint,
            'bias_constraint': bias_constraint,
            'data_format': data_format
        }

        return _args

class RNN(Layers):

    def __init__(self, gauge, layer_nm, rnn_args, rnn_type, state_out=0, mask_on=False, reset_type=None, **kwargs):

        self.gauge = gauge
        self.layer_nm = layer_nm

        super().__init__(gauge, layer_nm, **kwargs)

        self.rnn_args = rnn_args
        self.rnn_type = rnn_type

        self.state_out = state_out
        self.mask_on = mask_on
        self.reset_type = reset_type

        i_state_0 = tf.convert_to_tensor(np.ones((260,256), dtype=np.float64) * 3, dtype=tf.float64, name='i_state_0')
        i_state_1 = tf.convert_to_tensor(np.ones((260,256), dtype=np.float64) * 3, dtype=tf.float64, name='i_state_1')

        self.i_state_vars = [i_state_0,i_state_1]

        self.set_args(**self.rnn_args)

    def build(self, input_shp):

        self.input_shp = input_shp

        self.k_layer = self.get_rnn()

    def call(self, inputs, training=None, **kwargs):

        if self.reset_type is not None and self.stateful:
            self.reset_chk()

        if self.logs_on:

            state_0_pre = np_exp(self.k_layer.states[0])
            state_1_pre = np_exp(self.k_layer.states[1])

        if self.mask_on:
            z_outputs = self.k_layer(inputs=inputs, training=training, mask=self.gauge.masks['x_cols'])

        if not self.mask_on:
            z_outputs = self.k_layer(inputs=inputs, training=training)

        z_state = []

        if self.return_state:

            if len(z_outputs) == 2:
                z_state  = z_outputs[1]

            if len(z_outputs) == 3:
                z_state  = [z_outputs[1], z_outputs[2]]

            if self.state_out == 0:
                z_outputs = z_outputs[0]

        if self.logs_on:
            state_0_post = np_exp(self.k_layer.states[0])
            state_1_post = np_exp(self.k_layer.states[1])

        # -- logging -- #
        if self.logs_on:

            _specs =  {'epoch_idx': self.gauge.cur_run.epoch_idx.numpy(),
                       'batch_idx': self.gauge.cur_run.batch_idx.numpy()}

            _log = {
                'specs': _specs,
                'inputs': np_exp(inputs),
                'state_0_pre': state_0_pre,
                'state_1_pre': state_1_pre,
                'state_0_post': state_0_post,
                'state_1_post': state_1_post,
                'z_state_0': np_exp(z_state[0]),
                'z_state_1': np_exp(z_state[1]),
                'z_outputs': np_exp(z_outputs)}

            # self.logs.append(_log)
            self.save_layer_log(_log)

        return z_outputs

    def get_rnn(self):

        if self.rnn_type == 'gru':

            _rnn = tf_layers.GRU(units=self.units,
                                 activation=self.activation,
                                 recurrent_activation=self.recurrent_activation,
                                 use_bias=self.use_bias,
                                 kernel_initializer=self.kernel_initializer,
                                 recurrent_initializer=self.recurrent_initializer,
                                 bias_initializer=self.bias_initializer,
                                 kernel_regularizer=self.kernel_regularizer,
                                 recurrent_regularizer=self.recurrent_regularizer,
                                 bias_regularizer=self.bias_regularizer,
                                 kernel_constraint=self.kernel_constraint,
                                 recurrent_constraint=self.recurrent_constraint,
                                 bias_constraint=self.bias_constraint,
                                 dropout=self.dropout,
                                 recurrent_dropout=self.recurrent_dropout,
                                 implementation=self.implementation,
                                 reset_after=self.reset_after,
                                 return_sequences=self.return_sequences,
                                 return_state=self.return_state,
                                 go_backwards=self.go_backwards,
                                 stateful=self.stateful,
                                 unroll=self.unroll,
                                 time_major=self.time_major)

        if self.rnn_type == 'lstm':

            _rnn = tf_layers.LSTM(units=self.units,
                                  activation=self.activation,
                                  recurrent_activation=self.recurrent_activation,
                                  use_bias=self.use_bias,
                                  kernel_initializer=self.kernel_initializer,
                                  recurrent_initializer=self.recurrent_initializer,
                                  bias_initializer=self.bias_initializer,
                                  kernel_regularizer=self.kernel_regularizer,
                                  recurrent_regularizer=self.recurrent_regularizer,
                                  bias_regularizer=self.bias_regularizer,
                                  kernel_constraint=self.kernel_constraint,
                                  recurrent_constraint=self.recurrent_constraint,
                                  bias_constraint=self.bias_constraint,
                                  dropout=self.dropout,
                                  recurrent_dropout=self.recurrent_dropout,
                                  unit_forget_bias=self.unit_forget_bias,
                                  return_sequences=self.return_sequences,
                                  return_state=self.return_state,
                                  go_backwards=self.go_backwards,
                                  stateful=self.stateful,
                                  unroll=self.unroll,
                                  time_major=self.time_major)

        return _rnn

    def set_args(self, **kwargs):

        # -- default cell args -- #
        self.units = 1
        self.activation = 'tanh'
        self.recurrent_activation = 'sigmoid'
        self.use_bias = True
        self.kernel_initializer = 'glorot_uniform'
        self.recurrent_initializer = 'orthogonal'
        self.bias_initializer = 'zeros'
        self.kernel_regularizer = None
        self.recurrent_regularizer = None
        self.bias_regularizer = None
        self.kernel_constraint = None
        self.recurrent_constraint = None
        self.bias_constraint = None
        self.dropout = 0
        self.recurrent_dropout = 0

        self.reset_after = True
        self.unit_forget_bias = True

        # -- default rnn args -- #
        self.return_sequences = False
        self.return_state = False
        self.go_backwards = False
        self.stateful = False
        self.unroll = False
        self.time_major = False

        # -- cell args -- #

        if 'units' in kwargs:
            self.units = kwargs['units']

        if 'activation' in kwargs:
            self.activation = kwargs['activation']

        if 'recurrent_activation' in kwargs:
            self.recurrent_activation = kwargs['recurrent_activation']

        if 'use_bias' in kwargs:
            self.use_bias = kwargs['use_bias']

        if 'kernel_initializer' in kwargs:
            self.kernel_initializer = kwargs['kernel_initializer']

        if 'recurrent_initializer' in kwargs:
            self.recurrent_initializer = kwargs['recurrent_initializer']

        if 'bias_initializer' in kwargs:
            self.bias_initializer = kwargs['bias_initializer']

        if 'kernel_regularizer' in kwargs:
            self.kernel_regularizer = kwargs['kernel_regularizer']

        if 'recurrent_regularizer' in kwargs:
            self.recurrent_regularizer = kwargs['recurrent_regularizer']

        if 'bias_regularizer' in kwargs:
            self.bias_regularizer = kwargs['bias_regularizer']

        if 'kernel_constraint' in kwargs:
            self.kernel_constraint = kwargs['kernel_constraint']

        if 'recurrent_constraint' in kwargs:
            self.recurrent_constraint = kwargs['recurrent_constraint']

        if 'bias_constraint' in kwargs:
            self.bias_constraint = kwargs['bias_constraint']

        if 'dropout' in kwargs:
            self.dropout = kwargs['dropout']

        if 'recurrent_dropout' in kwargs:
            self.recurrent_dropout = kwargs['recurrent_dropout']

        if 'reset_after' in kwargs:
            self.reset_after = kwargs['reset_after']

        if 'unit_forget_bias' in kwargs:
            self.unit_forget_bias = kwargs['unit_forget_bias']

        # -- rnn args -- #

        if 'return_sequences' in kwargs:
            self.return_sequences = kwargs['return_sequences']

        if 'return_state' in kwargs:
            self.return_state = kwargs['return_state']

        if 'go_backwards' in kwargs:
            self.go_backwards = kwargs['go_backwards']

        if 'stateful' in kwargs:
            self.stateful = kwargs['stateful']

        if 'unroll' in kwargs:
            self.unroll = kwargs['unroll']

        if 'time_major' in kwargs:
            self.time_major = kwargs['time_major']

    def reset_chk(self):

        reset_type = self.reset_type

        is_new_epoch = self.gauge.is_new_epoch
        is_new_batch = self.gauge.is_new_batch
        is_new_day = self.gauge.is_new_day

        # print('-----', self.layer_nm, '------ \n')
        # print('New Epoch:', is_new_epoch)
        # print('New Batch:', is_new_batch)
        # print('New Day:', is_new_day)
        # print('\n\n')

        if reset_type == 'day' and is_new_day:
            self.k_layer.reset_states()
            tf.print(' ----------- DAY: reset state ----------- \n\n')
            return

        if reset_type == 'batch' and is_new_batch:
            self.k_layer.reset_states()
            tf.print(' ----------- BATCH: reset state ----------- \n\n')
            return

        if reset_type == 'epoch' and is_new_epoch:
            self.k_layer.reset_states()
            tf.print(' ----------- EPOCH: reset state ----------- \n\n')
            return

class Pool(Layers):

    def __init__(self,
                 gauge,
                 layer_nm,
                 pool_type='avg',
                 is_global=True,
                 pool_size=2,
                 strides=None,
                 tile=0,
                 padding='valid',
                 data_format='channels_last',
                 **kwargs):

        self.gauge = gauge
        self.layer_nm = layer_nm

        super().__init__(gauge, layer_nm, **kwargs)

        self.pool_type = pool_type
        self.is_global = is_global
        self.pool_size = pool_size
        self.strides = strides
        self.data_format = data_format
        self.tile = tile
        self.padding = padding

        if pool_type == 'avg':

            if is_global:
                self.k_layer = tf_layers.GlobalAveragePooling1D(name=self.layer_nm, data_format=self.data_format)

            if not is_global:
                self.k_layer = tf_layers.AveragePooling1D(name=self.layer_nm,
                                                         pool_size=self.pool_size,
                                                         strides=self.strides,
                                                         padding=self.padding,
                                                         data_format=self.data_format)

        if pool_type == 'max':

            if is_global:
                self.k_layer = tf_layers.GlobalMaxPool1D(name=self.layer_nm, data_format=self.data_format)

            if not is_global:
                self.k_layer = tf_layers.MaxPool1D(name=self.layer_nm,
                                                  pool_size=self.pool_size,
                                                  strides=self.strides,
                                                  padding=self.padding,
                                                  data_format=self.data_format)

    def call(self, inputs, training=None, **kwargs):

        z_output = self.k_layer(inputs=inputs, training=training)

        return z_output

class RunData(Layers):

    def __init__(self, gauge, layer_nm, rd_type, rd_args, **kwargs):

        self.rd_type = rd_type
        self.rd_args = rd_args

        self.rd_weights = 1.0
        self.rd_bias = 0.0

        super().__init__(gauge, layer_nm, **kwargs)

    def build(self, input_shp):

        self.input_shp = input_shp

        if self.rd_type.lower() == 'flat':

            if self.rd_args['weights_on']:
                self.rd_weights = self.add_weight(name=self.layer_nm + '_weights',
                                                      shape=(1,),
                                                      initializer=initializers.RandomUniform(minval=0., maxval=1.),
                                                      constraint=constraints.NonNeg(),
                                                      trainable=True)
            if self.rd_args['bias_on']:
                self.rd_bias = self.add_weight(name=self.layer_nm + '_bias',
                                               shape=1,
                                               initializer=initializers.Zeros(),
                                               constraint=constraints.NonNeg(),
                                               trainable=True)

        if self.rd_type.lower() == 'dnn':
            self.k_layer = tf_layers.Dense(name=self.layer_nm, **self.rd_args)

        if self.rd_type.lower() == 'cnn':
            self.k_layer = tf_layers.Conv1D(name=self.layer_nm, **self.rd_args)

        return

    def call(self, inputs, training=None, **kwargs):

        if not self.gauge.is_model_built:
            return inputs

        if self.gauge.is_model_built:

            if self.rd_type.lower() == 'flat':
                return tf.add(tf.multiply(inputs, self.rd_weights), self.rd_bias)

            if self.rd_type.lower() == 'dnn' or self.rd_type.lower() == 'cnn':
                return self.k_layer(inputs)


import tensorflow as tf
import numpy as np

from tensorflow.python.keras.utils import losses_utils
from tensorflow import math as tfm

# ::: Losses Class ::: #
class Losses():

    def __init__(self):

        self.MultiBCE = MultiBCE
        self.CatProb = CatProb

    def multi_loss(self, *args, **kwargs):
        return MultiLoss(*args, **kwargs)

    def multi_bce(self, *args, **kwargs):
        return MultiBCE(*args, **kwargs)

    def categorical_crossentropy(
            self,
            from_logits=False,
            label_smoothing=0,
            reduction=tf.keras.losses.Reduction.AUTO,
            name='photon_cc',
            **kwargs
        ):

        return tf.keras.losses.CategoricalCrossentropy(
            from_logits=from_logits,
            label_smoothing=label_smoothing,
            reduction=reduction,
            name=name
            )

    def binary_crossentropy(
            self,
            from_logits=False,
            label_smoothing=0,
            reduction=tf.keras.losses.Reduction.AUTO,
            name='photon_bc'
        ):

        return tf.keras.losses.BinaryCrossentropy(
            from_logits=from_logits,
            label_smoothing=label_smoothing,
            reduction=reduction,
            name=name
            )

    def mean_squared_error(self, reduction=losses_utils.ReductionV2.AUTO, name='photon_mse'):

        return tf.keras.losses.MeanSquaredError(reduction=reduction, name=name)

    def neg_log_like(self):

        return lambda y, p_y: -p_y.log_prob(y)

    def q_loss(
            self,
            **kwargs):

        return Qloss(**kwargs)

class NegLogLike():

    def __init__(self):
        pass

    def __call__(self, y_true, y_hat):

        z_loss = -y_hat.numpy().log_prob(y_true)

        return z_loss

class CatProb():

    def __init__(self):

        self.logs = []
        self.loss_fn = tf.keras.losses.categorical_crossentropy

    def __call__(self, y_true, y_hat, _run):

        z_loss = self.loss_fn(y_true, y_hat)

        # kl = sum(_run.chain.model.losses)

        # self.logs.append([z_loss,kl])

        # z_loss = z_loss + kl

        return z_loss

class MultiBCE():

    def __init__(self, n_funcs, loss_fn, loss_fn_args, logs_on=False, *args, **kwargs):

        self.n_funcs = n_funcs

        self.loss_fn = loss_fn
        self.loss_fn_args = loss_fn_args

        self.logs_on = logs_on

        self.args = args
        self.kwargs = kwargs

        self.loss_fns = []

        self.logs = []

        for i in range(self.n_funcs):
            self.loss_fns.insert(i, self.loss_fn(**self.loss_fn_args))

    def __call__(self, y_true, y_hat):

        z_losses = []

        z_true = []
        z_hat = []

        z_log = []

        for i in range(self.n_funcs):

            ed_idx = (i+1) * 2
            st_idx = ed_idx - 2

            _true = y_true[:,st_idx:ed_idx]
            _hat = y_hat[i]

            z_losses.insert(i, self.loss_fns[i](_true, _hat))

            z_true.append(_true)
            z_hat.append(_hat)

        reduced_losses = tf.reduce_sum(z_losses)

        if self.logs_on:

            _log = {'z_true': z_true,
                    'z_hat': z_hat,
                    'outputs': z_losses,
                    'reduced': reduced_losses}

            self.logs.append(_log)

        return [reduced_losses, z_true, z_hat]

class MultiLoss():

    def __init__(self, n_funcs, loss_fn, loss_fn_args, logs_on=False, *args, **kwargs):

        self.n_funcs = n_funcs

        self.loss_fn = loss_fn
        self.loss_fn_args = loss_fn_args

        self.logs_on = logs_on

        self.args = args
        self.kwargs = kwargs

        self.loss_fns = []

        self.logs = {'outputs': [], 'reduced': []}

        for i in range(self.n_funcs):
            self.loss_fns.insert(i, self.loss_fn(**self.loss_fn_args))

    def __call__(self, y_true, y_hat):

        z_losses = []

        for i in range(self.n_funcs):

            ed_idx = (i+1) * 2
            st_idx = ed_idx - 2

            # print(st_idx, ed_idx)

            _true = y_true[:,st_idx:ed_idx]
            _hat = y_hat[i]

            z_losses.insert(i, self.loss_fns[i](_true, _hat))

        reduced_losses = tf.reduce_sum(z_losses)

        if self.logs_on:
            self.logs['outputs'].append(z_losses)
            self.logs['reduced'].append(reduced_losses)

        return reduced_losses

class Qloss():

    def __init__(self,
                 from_logits=False,
                 label_smoothing=0,
                 reduction=tf.keras.losses.Reduction.AUTO,
                 logs_on=False):

        self.loss_args = {'from_logits': from_logits,
                          'label_smoothing': label_smoothing,
                          'reduction': reduction}

        self.loss_fn = tf.keras.losses.CategoricalCrossentropy(**self.loss_args)

        self.logs_on = logs_on
        self.logs = []

    def __call__(self, y_true, y_pred):

        cce_values = self.loss_fn(y_true, y_pred)

        y_prob = tf.nn.softmax(y_pred)

        neg_pens = tf.maximum(tf.subtract(y_true, y_prob),0)
        correct_pcts = tf.multiply(y_true, y_prob)
        fp_pens = tf.subtract(y_prob, correct_pcts)

        full_pens = tf.add(fp_pens, neg_pens)

        batch_size = y_true.shape[0]
        n_cols = y_true.shape[1]

        true_pos = tf.cast(tf.expand_dims(tf.math.argmax(y_true, axis=1), axis=1), dtype=tf.float64)
        pred_pos = tf.math.argmax(y_prob, axis=1)

        base_shp = (batch_size,n_cols)
        base_vals = tf.constant([[0,1,2,4,8,16,32]], dtype=tf.float64)
        pos_base = tf.reshape(tf.repeat(base_vals, repeats=batch_size, axis=0), shape=base_shp)

        pos_pens = tfm.abs(tfm.subtract(true_pos, pos_base))
        pos_norm = tfm.multiply(tfm.divide(pos_pens, n_cols),2)

        base_loss = tf.expand_dims(tf.reduce_sum(tfm.multiply(fp_pens, pos_norm), axis=1), axis=1)
        full_loss = tf.squeeze(tfm.add(base_loss, cce_values))

        if self.logs_on:

            _log = {'cce_values': cce_values,
                    'y_prob': y_prob,
                    'neg_pens': neg_pens,
                    'fp_pens': fp_pens,
                    'full_pens': full_pens,
                    'true_pos': true_pos,
                    'pred_pos': pred_pos,
                    'pos_base': pos_base,
                    'pos_pens': pos_pens,
                    'pos_norm': pos_norm,
                    'base_loss': base_loss,
                    'full_loss': full_loss}

            self.logs.append(_log)

        return tf.reduce_mean(full_loss)

class MultiBCE2():

    def __init__(self, reduction):

        self.reduction = reduction

        self.loss_fn = tf.keras.losses.BinaryCrossentropy()

    def __call__(self, y_true, y_hat):

        # --- 01 --- #

        l_01_true = y_true[:, :, 0:2]
        l_01_hat = y_hat[:, :, 0:2]

        s_01_true = y_true[:, :, 2:4]
        s_01_hat = y_hat[:, :, 2:4]

        l_01_loss = self.loss_fn(l_01_true, l_01_hat)
        s_01_loss = self.loss_fn(s_01_true, s_01_hat)

        # --- 02 --- #

        l_02_true = y_true[:, :, 4:6]
        l_02_hat = y_hat[:, :, 4:6]

        s_02_true = y_true[:, :, 6:8]
        s_02_hat = y_hat[:, :, 6:8]

        l_02_loss = self.loss_fn(l_02_true, l_02_hat)
        s_02_loss = self.loss_fn(s_02_true, s_02_hat)

        # --- 03 --- #

        l_03_true = y_true[:, :, 8:10]
        l_03_hat = y_hat[:, :, 8:10]

        s_03_true = y_true[:, :, 10:12]
        s_03_hat = y_hat[:, :, 10:12]

        l_03_loss = self.loss_fn(l_03_true, l_03_hat)
        s_03_loss = self.loss_fn(s_03_true, s_03_hat)

        # --- 04 --- #

        l_04_true = y_true[:, :, 12:14]
        l_04_hat = y_hat[:, :, 12:14]

        s_04_true = y_true[:, :, 14:16]
        s_04_hat = y_hat[:, :, 14:16]

        l_04_loss = self.loss_fn(l_04_true, l_04_hat)
        s_04_loss = self.loss_fn(s_04_true, s_04_hat)

        full_loss = tf.reduce_sum(
            l_01_loss + s_01_loss + l_02_loss + s_02_loss + l_03_loss + s_03_loss + l_04_loss +
            s_04_loss
            )

        return full_loss



import os, sys, asyncio, tracemalloc, threading, math, functools, time
import tensorflow as tf
import numpy as np
import pandas as pd

from concurrent import futures
from dataclasses import replace as dc_replace

def run_timer(func):
    @functools.wraps(func)
    def timer_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        call_func = func(*args, **kwargs)
        end_time = time.perf_counter()
        run_time = end_time - start_time

        run_units = 'secs'

        if (run_time > 60):
            run_time = run_time / 60
            run_units = 'mins'

        print(f'\n --- {run_time:.4f} {run_units} --- \n')

        return call_func

    return timer_wrapper

class Gamma():

    def __init__(self, network):

        self.network = network
        self.photon = self.network.photon

        if not self.photon.run_local:
            from photon.runs import setup_runs
            from photon.utils import dest_append, obj_exp
        else:
            sys.path.append(self.photon.mod_dir)
            from photon.runs import setup_runs
            from photon.utils import dest_append, obj_exp

        self.setup_runs = setup_runs
        self.dest_append = dest_append
        self.obj_exp = obj_exp

    @run_timer
    def run_network(self,
                    branches=None,
                    run_config=[],
                    epochs_on=True,
                    run_fn=None,
                    rebuild_on=False,
                    diag_on=False,
                    tf_log_on=False,
                    wandb_on=False):

        run = self.setup_runs(self.network, branches, run_config, run_fn, rebuild_on)

        if epochs_on:
            for branch_idx, branch in enumerate(run.branches):
                self.run_epochs(branch)

        return run

    def run_epochs(self, branch):

        # --- init msg --- #
        if branch.src.msgs_on:
            self.init_msg(branch.src)

        for epoch_idx in range(branch.src.n_epochs):

            is_on = False

            # --- loop chains --- #
            for chain_idx, chain in enumerate(branch.chains):

                if chain.config.run_fn != 'load':

                    if chain.config.run_type == 'infer' and epoch_idx >= 1:
                        break
                    else:
                        is_on = True

                        chain = self.update_live_chain(chain,
                                                       epoch_idx=epoch_idx,
                                                       run_type=chain.config.run_type,
                                                       data_type=chain.config.data_type,
                                                       is_val=False)
                        self.run_chain(chain)

            if is_on:
                self.save_chkps(branch)

                if branch.src.msgs_on:
                    self.epoch_msg(branch, epoch_idx)
                    self.save_log(branch, epoch_idx)

        return

    def run_chain(self, chain):

        chain.step_data_logs = []

        self.run_models(chain, is_val=False)

        # -- run val -- #
        if chain.config.val_on:
            self.run_models(chain, is_val=True)

        return

    def get_data(self, chain, is_val):

        data_type = chain.live.data_type

        dataset = chain.src.datasets[chain.live.tree_idx][data_type]

        for batch_idx, data in enumerate(dataset):

            inputs = data[0]
            targets = data[1]
            tracking = data[2]

            if not chain.src.data_config['targets']['is_seq']:
                targets = data[1][:, -1, :]

            batch_data = {'inputs': inputs,
                          'targets': targets,
                          'tracking': tracking}

            yield batch_idx, batch_data

    def get_models(self, chain, is_val):

        for model_idx, model in enumerate(chain.models):

            yield model_idx, model

    def run_models(self, chain, is_val):

        chain.specs = self.setup_specs(chain)

        if chain.config.async_on:
            self.run_models_async(chain, is_val)

        if not chain.config.async_on:
            for model_idx, model in self.get_models(chain, is_val):
                self.run_data(chain, model, model_idx, is_val, chain.specs['run_spec'][model_idx])

    def run_models_async(self, chain, is_val):

        chain.model_subs = []

        run_spec = chain.specs['run_spec']

        for r_spec in run_spec:

            chain.model_subs.append([])

            for d_spec in r_spec:

                async_idx = d_spec['async_idx']
                device_idx = d_spec['device_idx']
                start_idx = d_spec['start_idx']
                end_idx = d_spec['end_idx']
                max_wkrs = d_spec['max_wkrs']

                if start_idx is not None:

                    with futures.ThreadPoolExecutor(max_workers=max_wkrs) as model_exec:

                        for model_idx in range(start_idx, end_idx):

                            model = chain.models[model_idx]

                            model_sub = model_exec.submit(self.run_data, chain, model, model_idx, is_val, device_idx)
                            # model_sub.add_done_callback(self.async_callback_fn)

                            # async_sub = {'async_idx': async_idx,
                            #              'device_idx': device_idx,
                            #              'model_idx': model_idx,
                            #              'model_sub': model_sub}
                            #
                            # chain.model_subs[async_idx].insert(model_idx, async_sub)

                    futures.wait([async_sub['model_sub'] for async_sub in chain.model_subs[async_idx]])

    def async_callback_fn(self, _future):

        try:

            _future_res = _future.result()

        except AssertionError as err:

            print('Future Generated an Exception: %s' % (err))
            print('------------------------------- \n')

        # else:
        #
        #     print('CALL BACK')
            # self.save_steps(model_future_res['model'],
            #                 model_future_res['batch_idx'],
            #                 model_future_res['step_data'])

        return

    def run_data(self, chain, model, model_idx, is_val, device_idx):

        if not is_val:

            chain = self.update_live_chain(chain)
            model = self.update_live_model(model, chain=chain)

        # -- flip -- #
        if is_val:

            chain = self.update_live_chain(chain,
                                           run_type='val',
                                           data_type='val',
                                           is_val=True)

            model = self.update_live_model(model, chain)

        # -- loop batch data -- #
        for batch_idx, batch_data in self.get_data(chain, is_val):

            chain.live.batch_idx = batch_idx
            model.live.batch_idx = batch_idx

            self.run_steps(model, batch_data, batch_idx, device_idx)

        # -- flip back -- #
        if is_val:

            chain = self.update_live_chain(chain,
                                           run_type=chain.config.run_type,
                                           data_type=chain.config.data_type,
                                           is_val=False)

            model = self.update_live_model(model, chain)

        return

    def run_steps(self, model, batch_data, batch_idx, device_idx):

        model.gauge.is_live = True

        # ---- training steps ---- #
        if model.live.run_type == 'fit':

            model.live.is_training = True

            # --- load tape --- #
            with tf.GradientTape() as tape:

                with tf.device('/GPU:' + str(device_idx)):

                    step_data = model.src(inputs=batch_data['inputs'], training=True, batch_idx=batch_idx)
                    step_data = self.run_splits(model, step_data, batch_data)
                    step_data = self.run_loss(model, step_data)

                    # -- run grads -- #
                    step_data['step_grads'] = tape.gradient(step_data['step_loss'], model.src.trainable_variables)

                    # -- apply grads -- #
                    model.src.optimizer.apply_gradients(zip(step_data['step_grads'], model.src.trainable_variables))

                    # -- save lr -- #
                    step_data['learning_rates'] = model.src.optimizer.lr_sch.cur_lr

                    # model.chain.step_data_logs.append(step_data)

                    # ---- non-training steps ---- #
        if model.live.run_type in ['infer', 'pred', 'val']:

            model.live.is_training = False

            with tf.device('/GPU:' + str(device_idx)):

                step_data = model.src(inputs=batch_data['inputs'], training=True, batch_idx=batch_idx)
                step_data = self.run_splits(model, step_data, batch_data)
                step_data = self.run_loss(model, step_data)

                step_data['step_grads'] = None
                step_data['learning_rates'] = None

        # -- run metrics -- #
        if model.chain.config.metrics_on:

            n_metrics_fns = model.chain.config.n_metrics_fns

            step_data['metrics'] = np.ndarray(shape=[n_metrics_fns,])

            for idx in range(model.chain.config.n_metrics_fns):

                metrics_fn = model.gauge.metrics_fns[idx]

                step_data['metrics'][idx] = metrics_fn(step_data['y_true'], step_data['y_hat'], model.live.run_type)

        self.save_steps(model, step_data, batch_idx)

        return

    def run_splits(self, model, step_data, batch_data):

        targets_config = model.chain.src.data_config['targets']

        step_data['y_true'] = batch_data['targets'][targets_config['true_slice']]
        step_data['y_tracking'] = batch_data['targets'][targets_config['tracking_slice']]

        return step_data

    def run_loss(self, model, step_data):

        batch_size = model.chain.branch.src.trees[model.chain.live.tree_idx].data.batch_size

        # -- step loss -- #
        step_data['step_loss'] = model.gauge.loss_fn(step_data['y_true'], step_data['y_hat'])
        step_data['step_loss'] = tf.nn.compute_average_loss(step_data['step_loss'], global_batch_size=batch_size)

        # -- model loss -- #
        step_data['model_loss'] = sum(model.src.losses)

        # -- full loss -- #
        step_data['full_loss'] = tf.add(step_data['step_loss'], step_data['model_loss'])

        return step_data

    def update_live_chain(self, chain, **kwargs):

        for _key in kwargs.keys():
            setattr(chain.live, _key, kwargs[_key])

        return chain

    def update_live_model(self, model, chain=None, **kwargs):

        if chain is not None:

            model.live.batch_idx = chain.live.batch_idx
            model.live.epoch_idx = chain.live.epoch_idx
            model.live.data_type = chain.live.data_type
            model.live.run_type = chain.live.run_type
            model.live.is_val = chain.live.is_val

        for _key in kwargs.keys():
            setattr(model.live, _key, kwargs[_key])

        return model

    def setup_specs(self, chain):

        n_devices = chain.src.photon.n_gpus
        n_models = chain.n_models

        device_dist = [80, 10, 10]

        if chain.config.async_on:

            total_wrks = sum([device_dist[idx] for idx in range(n_devices)])

            n_async_calls =  math.ceil(n_models / total_wrks)

            device_spec = []

            for device_idx in range(n_devices):

                max_wkrs = device_dist[device_idx]

                max_wkrs_pct = max_wkrs / total_wrks

                max_models = math.floor(n_models * max_wkrs_pct)

                _spec = {'device_idx': device_idx,
                         'max_wkrs': max_wkrs,
                         'max_wkrs_pct': max_wkrs_pct,
                         'max_models': max_models,
                         'n_async_calls': n_async_calls}

                device_spec.insert(device_idx, _spec)

            run_spec = []

            _end_idx = 0

            for async_idx in range(n_async_calls):

                run_spec.append([])

                scale = async_idx

                for device_idx in range(n_devices):

                    max_wkrs = device_spec[device_idx]['max_wkrs']

                    _end_idx += max_wkrs

                    _start_idx = _end_idx - max_wkrs

                    spec_models = []

                    if _start_idx < n_models:

                        start_idx = _start_idx

                        if _end_idx <= n_models:
                            end_idx = _end_idx
                        else:
                            end_idx = n_models

                    if _start_idx >= n_models:

                        start_idx = None
                        end_idx = None

                    _spec = {'async_idx': async_idx,
                             'device_idx': device_idx,
                             'max_wkrs': max_wkrs,
                             '_start_idx': _start_idx,
                             '_end_idx': _end_idx,
                             'start_idx': start_idx,
                             'end_idx': end_idx}

                    run_spec[async_idx].insert(device_idx, _spec)

        if not chain.config.async_on:

            device_calls = sum([device_dist[idx] for idx in range(n_devices)])

            device_spec = []
            run_spec = []

            model_idx = 0

            for device_idx in range(n_devices):

                device_models = math.ceil(n_models * (device_dist[device_idx] / device_calls))

                _spec = {'device_idx': device_idx,
                         'device_calls': device_calls,
                         'device_models': device_models}

                device_spec.insert(device_idx, _spec)

                for n in range(device_models):

                    if model_idx <= n_models:

                        run_spec.insert(model_idx, device_idx)

                    model_idx += 1

        return {'device_spec': device_spec,
                'run_spec': run_spec}

    def save_steps(self, model, step_data, batch_idx):

        save_config = model.chain.src.save_config.copy()

        if model.live.run_type != 'fit':
            del (save_config['grads'])
            del (save_config['learning_rates'])

        if not model.chain.config.metrics_on:
            del (save_config['metrics'])

        _keys = step_data.keys()

        for _key in _keys:

            output_epoch_idx = model.live.epoch_idx

            if _key in save_config and save_config[_key] is not None:

                if save_config[_key] == 'last':
                    output_epoch_idx = 0

                if _key not in ['step_loss', 'model_loss', 'full_loss']:
                    _obj = np.asarray(self.obj_exp(step_data[_key], to_numpy=True, strat=model.gauge.strat, axis=0))

                if _key in ['step_loss', 'model_loss', 'full_loss']:
                    _obj = np.asarray(self.obj_exp(step_data[_key], strat=model.gauge.strat, reduce='sum', axis=None))

                if model.live.is_val:
                    self.dest_append(_obj,
                                     model.val_steps,
                                     key=_key,
                                     epoch_idx=output_epoch_idx,
                                     batch_idx=batch_idx)

                if not model.live.is_val:
                    self.dest_append(_obj,
                                     model.steps,
                                     key=_key,
                                     epoch_idx=output_epoch_idx,
                                     batch_idx=batch_idx)

        # -- turn off gauge -- #
        model.gauge.is_live = False

        # -- increment step_idx -- #
        if model.live.run_type == 'fit':
            model.live.step_idx.assign_add(1)

        return

    def save_chkps(self, branch):

        for chain_idx, chain in enumerate(branch.chains):

            if chain.config.save_cp and chain.live.run_type == 'fit':

                for model in chain.models:

                    check_interval = True

                    if chain.live.epoch_idx == branch.src.n_epochs-1:
                        check_interval = False

                    model.gauge.chkp_manager.save(check_interval=check_interval)

    # ------ msgs ------ #

    def init_msg(self, branch):

        tree_idx = 0

        photon = branch.photon
        photon_id = branch.photon.photon_id

        data_type = 'train'
        n_epochs = branch.n_epochs
        n_chains = branch.n_chains

        samples_pd = branch.trees[tree_idx].data.samples_pd

        batch_size = branch.trees[tree_idx].data.store[data_type]['config']['batch_size']
        n_batches = branch.trees[tree_idx].data.store[data_type]['config']['n_batches']
        n_calls = branch.trees[tree_idx].data.store[data_type]['config']['n_calls']
        n_steps = branch.trees[tree_idx].data.store[data_type]['config']['n_steps']

        train_days = branch.trees[tree_idx].data.store['train']['config']['n_days']
        test_days = branch.trees[tree_idx].data.store['test']['config']['n_days']
        val_days = branch.trees[tree_idx].data.store['val']['config']['n_days']

        seq_len = branch.trees[tree_idx].data.seq_len

        msg = f'\n'

        msg += f' ------------------- {branch.network.name} -------------------'

        msg += f'\n\n'

        msg += f'   Photon ID: {photon_id} |'
        msg += f' Tree: {branch.trees[tree_idx].name} |'
        msg += f' Branch: {branch.name} '

        msg += f'\n\n'
        msg += f'   Epochs: {n_epochs} |'
        msg += f' Chains: {n_chains} |'
        msg += f' Batch Size: {batch_size} '

        msg += f'\n\n'
        msg += f'   Train Days: {train_days} |'
        msg += f' Val Days: {val_days} |'
        msg += f' Test Days: {test_days}'

        msg += f'\n\n'

        msg += f' -------------------------------------------------------'

        tf.print(msg)

    def epoch_msg(self, branch, epoch_idx):

        epoch_idx = epoch_idx
        loss_epoch_idx = epoch_idx

        branch.epoch_msg = {'data': [],
                            'header': f"",
                            'body': f"",
                            'lc_body': f"",
                            'footer': f""}

        # --- header --- #

        branch.epoch_msg['header'] += f"\n ****************** Epoch: {epoch_idx + 1} ****************** "

        # --- loop chains --- #
        for chain_idx, chain in enumerate(branch.chains):

            chain_lr = 0

            if chain.live.run_type == 'fit':
                chain_lr = np.mean(chain.models[0].steps.learning_rates[epoch_idx])

            branch.epoch_msg['data'].append({'chain_lr': chain_lr,
                                             'main_loss': [],
                                             'main_acc': {'models_acc': [], 'all_acc': None, 'per_acc_batches': [], 'per_acc': []},
                                             'val_loss': [],
                                             'val_acc': {'models_acc': [], 'all_acc': None, 'per_acc_batches': [], 'per_acc': []}})

            # --- loop models --- #

            for model_idx, model in enumerate(chain.models):

                # ---- build body components ---- #

                if model.live.run_type == 'infer':

                    if epoch_idx > 0:
                        loss_epoch_idx = 0

                branch.epoch_msg['data'][chain_idx]['main_loss'].insert(model_idx, np.mean(model.steps.full_loss[loss_epoch_idx]))

                if chain.config.val_on:

                    branch.epoch_msg['data'][chain_idx]['val_loss'].insert(model_idx, np.mean(model.val_steps.full_loss[loss_epoch_idx]))

                if chain.config.metrics_on:

                    branch.epoch_msg['data'][chain_idx]['main_acc']['models_acc'].insert(model_idx, np.asarray(model.steps.metrics)[loss_epoch_idx])

                    if chain.config.val_on:

                        branch.epoch_msg['data'][chain_idx]['val_acc']['models_acc'].insert(model_idx, np.asarray(model.val_steps.metrics)[loss_epoch_idx])

            branch.epoch_msg['data'][chain_idx]['main_acc']['all_acc'] = np.asarray(branch.epoch_msg['data'][chain_idx]['main_acc']['models_acc'])
            branch.epoch_msg['data'][chain_idx]['val_acc']['all_acc'] = np.asarray(branch.epoch_msg['data'][chain_idx]['val_acc']['models_acc'])

            # -- loop metrics fns -- #

            for idx in range(chain.config.n_metrics_fns):

                main_per_acc = branch.epoch_msg['data'][chain_idx]['main_acc']['all_acc'][:, :, idx]
                val_per_acc = branch.epoch_msg['data'][chain_idx]['val_acc']['all_acc'][:, :, idx]

                branch.epoch_msg['data'][chain_idx]['main_acc']['per_acc_batches'].insert(idx, main_per_acc)
                branch.epoch_msg['data'][chain_idx]['main_acc']['per_acc'].insert(idx, main_per_acc[:, -1])

                branch.epoch_msg['data'][chain_idx]['val_acc']['per_acc_batches'].insert(idx, val_per_acc)
                branch.epoch_msg['data'][chain_idx]['val_acc']['per_acc'].insert(idx, val_per_acc[:, -1])

            # --- build body --- #

            body_msg = self.epoch_msg_body(branch,chain,chain_idx,chain_lr)

            branch.epoch_msg['body'] += body_msg

            if chain_idx == branch.src.n_chains-1:
                branch.epoch_msg['lc_body'] += body_msg

        tf.print(branch.epoch_msg['header'])
        tf.print(branch.epoch_msg['body'])
        tf.print(branch.epoch_msg['footer'])

    def epoch_msg_body(self, branch, chain, chain_idx, chain_lr):
        
        body_msg = f""

        if chain_idx > 0:
            body_msg += f"\n"

        body_msg += f"\n    ::::: {chain.src.name} ({chain.n_models} models) ::::: \n"

        type_hdr = chain.live.run_type.upper()

        if type_hdr == 'FIT':
            type_hdr = 'TRAIN'

        main_loss = branch.epoch_msg['data'][chain_idx]['main_loss']

        body_msg += f"\n\t\t -------- {type_hdr} -------- \n"

        if chain.live.run_type == 'fit':
            body_msg += f"\n\t\t"
            body_msg += f" ::: Avg LR {chain_lr:.7f} :::"
            body_msg += f"\n\n\t\t"

        if chain.live.run_type != 'fit':
            body_msg += f"\n\t\t"

        body_msg += f" Loss \t {np.mean(main_loss):.3f}"

        if chain.n_models > 1:
            body_msg += f" ({np.std(main_loss):.3f})"

        if chain.config.metrics_on:

            main_per_acc = np.asarray(branch.epoch_msg['data'][chain_idx]['main_acc']['per_acc'])

            body_msg += f"\n\t\t"
            body_msg += f" Acc \t"

            for idx in range(chain.config.n_metrics_fns):
                main_acc = main_per_acc[idx]

                if idx > 0:
                    body_msg += f"\n\t\t\t\t"

                body_msg += f" {np.mean(main_acc):.3f}"
                if chain.n_models > 1:
                    body_msg += f" ({np.std(main_acc):.3f})"

        # --- val --- #

        if chain.config.val_on:

            val_loss = branch.epoch_msg['data'][chain_idx]['val_loss']

            body_msg += f"\n"
            body_msg += f"\n\t\t -------- VAL -------- \n"
            body_msg += f"\n\t\t"
            body_msg += f" Loss \t {np.mean(val_loss):.3f}"
            if chain.n_models > 1:
                body_msg += f" ({np.std(val_loss):.3f})"

            # -- loop metrics fns -- #

            if chain.config.metrics_on:
                val_per_acc = np.asarray(branch.epoch_msg['data'][chain_idx]['val_acc']['per_acc'])

                body_msg += f"\n\t\t"
                body_msg += f" Acc \t"

                for idx in range(chain.config.n_metrics_fns):

                    val_acc = val_per_acc[idx]

                    if idx > 0:
                        body_msg += f"\n\t\t\t\t"

                    body_msg += f" {np.mean(val_acc):.3f}"
                    if chain.n_models > 1:
                        body_msg += f" ({np.std(val_acc):.3f})"

        return body_msg

    def save_log(self, branch, epoch_idx):

        log_msg = f"\n {branch.epoch_msg['header']} \n"
        log_msg += f"{branch.epoch_msg['body']} \n"

        lc_log_msg = f"\n {branch.epoch_msg['header']} \n"
        lc_log_msg += f"{branch.epoch_msg['lc_body']} \n"

        # --- photon log --- #
        if branch.run.network.msgs_on['photon_log']:

            run_id = str(branch.run.id)
            run_idx = str(branch.run.network.n_runs)

            runs_dir = branch.src.photon.store['runs'] + '/' + run_id

            if not os.path.exists(runs_dir):
                os.makedirs(runs_dir)

            runs_file_path = runs_dir + '/run_' + run_idx + '.log'
            runs_lc_file_path = runs_dir + '/run_lc_' + run_idx + '.log'

            # --- write run file --- #
            with open(runs_file_path, "a") as runs_file:
                runs_file.write(log_msg)

            # --- write last chain file --- #
            with open(runs_lc_file_path, "a") as runs_lc_file:
                runs_lc_file.write(lc_log_msg)


import sys

from typing import List, Tuple, Dict, Optional
from photon import Photon, layers

from photon.utils import np_exp

import tensorflow as tf
from tensorflow import keras as tfk
from tensorflow.keras import layers as tfkl

import numpy as np
import pandas as pd

from tensorflow.keras import layers as tf_layers
from tensorflow.keras import Model as tf_Model
from tensorflow.keras import activations, initializers, regularizers, constraints

from tensorflow.python.keras.utils import losses_utils

from sklearn import preprocessing

class Models(tf_Model):

    def __init__(self, **kwargs):

        self.gauge = kwargs['gauge']

        super().__init__(name=self.gauge.name)

        self.inputs_ph = None

        self.d_model = self.gauge.model_args['d_model']
        self.reg_args = self.gauge.model_args['reg_args']
        self.norm_args = self.gauge.model_args['norm_args']
        self.drop_rate = self.gauge.model_args['reg_vals'][0]
        self.seed = self.gauge.model_args['seed']
        self.show_calls = self.gauge.model_args['show_calls']

        self.model_idx = self.gauge.model_idx
        self.chain = self.gauge.chain
        self.chain_idx = self.chain.chain_idx
        self.branch = self.chain.branch
        self.branch_idx = self.branch.branch_idx

        self.epoch_idx = None
        self.batch_idx = None
        self.run_type = None
        self.data_type = None
        self.is_val = None
        self.is_training = None

        self.layer_logs = []
        self.rd_logs = []

        self.reg_layers = {}
        self.norm_layers = {}

        self.z_return = {}

    def __call__(self, inputs, training, **kwargs):

        if 'epoch_idx' in kwargs:
            self.epoch_idx = kwargs['epoch_idx']

        if 'batch_idx' in kwargs:
            self.batch_idx = kwargs['batch_idx']

        if 'run_type' in kwargs:
            self.run_type = kwargs['run_type']

        if 'data_type' in kwargs:
            self.data_type = kwargs['data_type']

        if 'is_val' in kwargs:
            self.is_val = kwargs['is_val']

        if 'is_training' in kwargs:
            self.is_training = kwargs['is_training']

        if not self.gauge.is_built:
            self.build_model()

        if self.gauge.is_model_built:

            if self.show_calls:
                tf.print( '\nrun_type:', self.gauge.run_chain.live.run_type,

                          '| chain_idx:', self.gauge.run_chain.src.chain_idx,
                          '| model_idx:', self.gauge.model_idx,

                          '| epoch_idx:', self.gauge.run_model.live.epoch_idx,
                          '| batch_idx:', self.gauge.run_model.live.batch_idx,

                          '| run_type:', self.gauge.run_model.live.run_type,
                          '| data_type:', self.gauge.run_model.live.data_type,
                          '| is_val:', self.gauge.run_model.live.is_val,
                          '| is_training:', self.gauge.run_model.live.is_training,

                          output_stream=sys.stdout, sep=' ', end='\n')

            return self.call(inputs)

    def pre_build(self, input_data):

        self.build_model()

        self.build(input_data.shape)

        self.gauge.is_model_built = True

    def build_run_data(self, config):

        '''

            args:

                branch_idx: The idx of the branch to use as the rd_branch

                    0 = Sets it to the same branch.

                    n<0 = Sets the rd_branch to n branches before the current branch.
                            Example: -1 sets it to the previous branch.

                    n>0 = Sets the rd_branch to that of n.

                chain_idx: The idx of the chain to use as the rd_chain

                    0 = Sets it to the same chain.

                    n<0 = Sets the rd_chain to n chains before the current chain.
                            Example: -1 makes it the previous chain.

                    n>0 = Sets the rd_chain to that of n.

                n_inputs: Used as the input shape of the inputs placeholder.

                    0 = Will make it default to the number of models in the rd_chain.

                    n>0 = The number of inputs for the placeholder shape

                n_layers: Statically set the input shape of the inputs placeholder

                    0 = Will make it default to the number of models in the rd_chain. This makes
                        it a 1:1 releasionship between the number of models in the rd_chain to
                        the number of rd layers.

                    n>0 = Statically set the number of rd layers. Normally would just use 1 to make
                            make just one rd layer for a many to one setup.

                rd_idx_type: Used when calling the rd layers.

                    layer_idx: matches the layer_idx of the rd_layer
                    model_idx: uses the current model_idx

                    ** if n_layers = 1 and rd_idx_type = layer_idx all rd inputs will come from pre chain model 0
                    because it it will match the rdx of the zero indexed 1 layer

                    ** if n_layers = 1 and rd_idx_type = model_idx rd inputs will match the current model_idx to the
                    model_idx from the previous chain

                rd_type: Set the type of the rd layers
                    'flat' =
                    'dnn' =
                    'cnn' =

                ** if layer is not flat the units/filters need to match the rd_n_inputs **
        '''

        self.rd_layers = []

        self.rd_branch_idx = self.branch_idx + config['branch_idx']
        self.rd_chain_idx = self.chain_idx + config['chain_idx']
        self.rd_n_inputs = config['n_inputs']
        self.rd_n_layers = config['n_layers']
        self.rd_type = config['type']
        self.rd_args = config['args']
        self.rd_idx_type = config['rd_idx_type']
        self.pass_on = config['pass_on']
        self.rd_trans_on = config['trans_on']
        self.rd_squeeze_on = config['squeeze_on']
        self.rd_softmax_on = config['softmax_on']
        
        self.rd_logs_on = config['logs_on']

        # -- set the rd_chain -- #
        self.rd_chain = self.gauge.runs[-1].branches[self.rd_branch_idx].chains[self.rd_chain_idx]

        # -- override rd_n_inputs -- #
        if self.rd_n_inputs == 0:
            self.rd_n_inputs = self.rd_chain.n_models

        # -- setup n_layers -- #
        if self.rd_n_layers == 0:
            self.rd_n_layers = self.rd_chain.n_models

        for layer_idx in range(self.rd_n_layers):
            self.rd_layers.append(layers.RunData(self.gauge,
                                                 layer_nm=self.gauge.name + '_rd_' + str(layer_idx),
                                                 rd_type=self.rd_type,
                                                 rd_args=self.rd_args,
                                                 reg_args=None,
                                                 norm_args=None))

        return
    
    def call_run_data(self, inputs):
        
        batch_size = self.chain.trees[0].data.batch_size
        dtype = self.chain.trees[0].data.dtype

        epoch_idx = self.gauge.run_model.live.epoch_idx
        batch_idx = self.gauge.run_model.live.batch_idx

        rd_outputs = []

        log_data = {'idx': [],
                    'inputs': [],
                    'outputs': [],
                    'weights': []}

        # -- output rd layers -- #
        for _idx in range(self.rd_n_layers):

            if self.rd_idx_type == 'layer_idx':
                rd_idx = _idx

            if self.rd_idx_type == 'model_idx':
                rd_idx = self.model_idx

            # -- if not built -- #
            if not self.gauge.is_model_built:

                self.inputs_ph = tfk.Input(shape=(self.rd_n_inputs,),
                                           batch_size=batch_size,
                                           dtype=dtype,
                                           name='input_data_' + str(_idx))

                # -- call rd layer -- #
                rd_outputs.insert(_idx, self.rd_layers[_idx](self.inputs_ph))

            # -- if built -- #
            if self.gauge.is_model_built:

                if self.gauge.run_model.live.run_type != 'val':
                    pre_steps = self.rd_chain.models[rd_idx].steps

                if self.gauge.run_model.live.run_type == 'val':
                    pre_steps = self.rd_chain.models[rd_idx].val_steps

                # -- create inputs for passthrough or to send to rd_layers -- #
                _inputs = tf.convert_to_tensor(pre_steps.y_hat[-1][batch_idx])

                # -- passthrough or call rd_layers -- #
                if self.pass_on:
                    _outputs = _inputs

                else:
                    _outputs = self.rd_layers[_idx](_inputs)

                # -- save into rd_outputs regardless if passthrough or rd_layers called -- #
                rd_outputs.insert(_idx, _outputs)

                # -- logs -- #
                if self.rd_logs_on and not self.gauge.run_model.live.is_val:

                    log_data['idx'].insert(_idx, _idx)
                    log_data['inputs'].insert(_idx, _inputs)
                    log_data['outputs'].insert(_idx, _outputs)
                    log_data['weights'].insert(_idx, self.trainable_variables[0].numpy())

        z_outputs = tf.convert_to_tensor(rd_outputs)

        # --- transpose output --- #
        if self.rd_trans_on:
            z_outputs = tf.transpose(z_outputs, perm=[1, 0, 2])

        # --- squeeze output --- #
        if self.rd_squeeze_on:
            z_outputs = tf.squeeze(z_outputs)

        # --- squeeze output --- #
        if self.rd_softmax_on:
            z_outputs = tf.nn.softmax(z_outputs)

        # --- logs --- #
        if self.gauge.is_model_built and self.rd_logs_on and not self.gauge.run_model.live.is_val:

            if len(self.rd_logs) <= epoch_idx:
                self.rd_logs.append([])

            _log = {'layers': log_data, 'z_outputs': z_outputs}

            self.rd_logs[epoch_idx].insert(batch_idx, _log)

        return z_outputs


import os, sys, math, pickle, json, shelve, struct, pathlib

from dataclasses import dataclass, field, replace as dc_replace
from typing import List, Dict, Any

from contextlib import redirect_stdout, nullcontext

from concurrent.futures import ThreadPoolExecutor

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

import tensorflow as tf

from tensorflow.keras import backend as K

from tensorflow import distribute as tf_dist

from sklearn import preprocessing

class Photon():

    def __init__(self,
                 run_local=False,
                 run_dir=None,
                 cuda_on=True,
                 limit_mem=True):

        self.cuda_on = cuda_on
        self.limit_mem = limit_mem

        self.run_local = run_local
        self.run_dir = run_dir
        self.mod_dir = pathlib.Path(__file__).parent.parent

        if not self.run_local:
            from photon import metrics, losses
            from photon.gamma import Gamma
        else:
            sys.path.append(self.mod_dir)
            from photon import metrics, losses
            from photon.gamma import Gamma

        self.Networks = Networks
        self.Trees = Trees
        self.Branches = Branches
        self.Chains = Chains
        self.Gamma = Gamma

        self.metrics = metrics.Metrics()
        self.losses = losses.Losses()

        self.n_gpus = 0
        self.n_v_gpus = 0

        if self.cuda_on:

            self.physical_cpus = tf.config.list_physical_devices('CPU')
            self.physical_gpus = tf.config.list_physical_devices('GPU')

            self.n_gpus = len(self.physical_gpus)

            if self.limit_mem:
                for d in self.physical_gpus:
                    tf.config.experimental.set_memory_growth(d, True)

        self.set_options(2)

    def set_options(self, fp):

        max_rows = 500
        max_cols = 25
        edgeitems = 30
        linewidth = 900
        threshold = 10000
        all_cols = False

        float_p = '%.' + str(fp) + 'f'

        ff = lambda x: float_p % x

        pd.options.display.float_format = ff
        pd.set_option('display.max_rows', max_rows)
        pd.set_option('display.max_columns', max_cols)

        if all_cols:
            pd.set_option('display.expand_frame_repr', False)

        np.set_printoptions(
            formatter={'float': ff}, edgeitems=edgeitems, linewidth=linewidth, threshold=threshold
        )

    def setup_photon(self, photon_load_id):

        if self.run_dir is None:
            self.run_dir = os.path.expanduser('~') + '/photon_temp'

        self.store_dir = self.run_dir + '/store'

        self.store = {'db': self.store_dir + '/db',
                      'runs': self.store_dir + '/runs',
                      'chkps': self.store_dir + '/chkps'}

        if not os.path.exists(self.store_dir):
            os.makedirs(self.store_dir)
            os.makedirs(self.store['db'])
            os.makedirs(self.store['runs'])
            os.makedirs(self.store['chkps'])

        db = shelve.open(self.store['db']+'/db')

        if photon_load_id == 0:

            if 'photon_id' in db:
                db['photon_id']+=1
            else:
                db['photon_id'] = 1

            self.photon_id = db['photon_id']

        if photon_load_id > 0:

            self.photon_id = photon_load_id

        self.photon_nm = 'Photon_' + str(self.photon_id)

        db.close()

        return

class Networks():

    def __init__(self,
                 photon_id,
                 data_dir,
                 data_fn,
                 data_res,
                 data_cols,
                 x_groups_on,
                 dirs_on,
                 diag_on,
                 msgs_on,
                 name=None,
                 photon=None,
                 float_x=32):

        self.is_built = False

        self.photon_load_id = photon_id

        if name is None:
            self.name = 'Photon Network'
        else:
            self.name = name

        if photon is None:
            self.photon = Photon()
        else:
            self.photon = photon

        self.photon.setup_photon(self.photon_load_id)

        self.gamma = self.photon.Gamma(self)

        self.x_groups_on = x_groups_on

        self.dirs_on = dirs_on

        self.data = None
        self.data_dir = data_dir
        self.data_fn = data_fn
        self.data_res = data_res
        self.data_cols = data_cols

        self.diag_on = diag_on
        self.msgs_on = msgs_on

        self.msgs_on['epochs'] = False

        if float_x == 16:
            self.float_x = 'float16'
            self.dtype = np.float16

        if float_x == 32:
            self.float_x = 'float32'
            self.dtype = np.float32

        if float_x == 64:
            self.float_x = 'float64'
            self.dtype = np.float64

        # -- load data -- #
        self.load_data()

        self.n_trees = 0
        self.n_branches = 0
        self.n_chains = 0
        self.n_runs = 0

        self.trees = []
        self.branches = []
        self.chains = []
        self.runs = []

        K.set_floatx(self.float_x)

    def add_tree(self, tree):

        self.trees.append(tree)
        self.n_trees += 1

        return self.n_trees - 1, self.data

    def add_branch(self, branch):

        self.branches.append(branch)
        self.n_branches += 1

        return self.n_branches - 1

    def load_data(self):

        self.data_fp = self.data_dir + '/' + self.data_fn + '.parquet'

        self.data = self.Data()

        self.setup_cols(self.data_cols)

        data_tab = pq.read_table(self.data_fp)

        self.data.full_bars = data_tab.to_pandas().astype(self.dtype)

        all_cols = self.data.all_cols + ['is_close']

        self.data.full_bars = self.data.full_bars[all_cols]

        return

    def setup_cols(self, data_cols):

        # ------- x_cols ------- #

        self.data.x_cols = list(data_cols['x_cols'].keys())

        # ------- c_cols ------- #

        if data_cols['c_cols'] is not None:
            self.data.c_cols = list(data_cols['c_cols'].keys())

        if data_cols['c_cols'] is None:
            self.data.c_cols = []

        # ------- y_cols ------- #

        if data_cols['y_cols'] is not None:
            self.data.y_cols = list(data_cols['y_cols'].keys())

        if data_cols['y_cols'] is None:
            self.data.y_cols = []

        # ------- t_cols ------- #

        if data_cols['t_cols'] is not None:
            self.data.t_cols = list(data_cols['t_cols'].keys())

        if data_cols['t_cols'] is None:
            self.data.t_cols = []

        self.data.n_x_cols = len(self.data.x_cols)
        self.data.n_c_cols = len(self.data.c_cols)
        self.data.n_y_cols = len(self.data.y_cols)
        self.data.n_t_cols = len(self.data.t_cols)

        # -- x cols -- #
        for i in range(self.data.n_x_cols):

            _col = self.data.x_cols[i]
            _base = data_cols['x_cols'][_col]

            self.data.agg_data[_col] = _base['seq_agg']

            if _base['ofs_on']:
                self.data.ofs_data['x'].append(_col)

            if not self.x_groups_on:
                if _base['nor_on']:
                    self.data.nor_data['x'].append(_col)

        # -- c cols -- #
        if data_cols['c_cols'] is not None:

            for i in range(self.data.n_c_cols):

                _col = self.data.c_cols[i]
                _base = data_cols['c_cols'][_col]

                self.data.agg_data[_col] = _base['seq_agg']

                if _base['ofs_on']:
                    self.data.ofs_data['c'].append(_col)

                if _base['nor_on']:
                    self.data.nor_data['c'].append(_col)

        # -- y cols -- #
        if data_cols['y_cols'] is not None:

            for i in range(self.data.n_y_cols):

                _col = self.data.y_cols[i]
                _base = data_cols['y_cols'][_col]

                self.data.agg_data[_col] = _base['seq_agg']

                if _base['ofs_on']:
                    self.data.ofs_data['y'].append(_col)

                if _base['nor_on']:
                    self.data.nor_data['y'].append(_col)

        # -- t cols -- #
        for i in range(self.data.n_t_cols):

            _col = self.data.t_cols[i]
            _base = data_cols['t_cols'][_col]

            self.data.agg_data[_col] = _base['seq_agg']

            if _base['ofs_on']:
                self.data.ofs_data['t'].append(_col)

            if _base['nor_on']:
                self.data.nor_data['t'].append(_col)

        if self.x_groups_on:
            self.data.x_groups = self.setup_x_groups(data_cols['x_cols'])

        self.data.close_cols = ['bar_idx',
                                'day_idx',
                                'BAR_TP']

        self.data.all_cols = self.data.x_cols + self.data.c_cols + self.data.y_cols + self.data.t_cols

    def setup_x_groups(self, x_cols):

        pr_group = []

        for k, v in x_cols.items():

            for k2, v2 in v.items():

                if k2 == 'x_group' and v2 == 'pr':
                    pr_group.append(k)

        vol_group = []

        for k, v in x_cols.items():

            for k2, v2 in v.items():

                if k2 == 'x_group' and v2 == 'vol':
                    vol_group.append(k)

        atr_group = []

        for k, v in x_cols.items():

            for k2, v2 in v.items():

                if k2 == 'x_group' and v2 == 'atr':
                    atr_group.append(k)

        roc_group = []

        for k, v in x_cols.items():

            for k2, v2 in v.items():

                if k2 == 'x_group' and v2 == 'roc':
                    roc_group.append(k)

        zsc_group = []

        for k, v in x_cols.items():

            for k2, v2 in v.items():

                if k2 == 'x_group' and v2 == 'zsc':
                    zsc_group.append(k)

        return {'pr_group': pr_group,
                'vol_group': vol_group,
                'atr_group': atr_group,
                'roc_group': roc_group,
                'zsc_group': zsc_group}

    @dataclass
    class Data:

        x_cols: List = field(default_factory=lambda: [[]])
        c_cols: List = field(default_factory=lambda: [[]])
        y_cols: List = field(default_factory=lambda: [[]])
        t_cols: List = field(default_factory=lambda: [[]])

        close_cols: List = field
        all_cols: List = field(default_factory=lambda: [[]])

        n_x_cols: List = 0
        n_c_cols: List = 0
        n_y_cols: List = 0
        n_t_cols: List = 0

        agg_data: Dict = field(default_factory=lambda: {})

        ofs_data: List = field(default_factory=lambda: {
            'x': [],
            'c': [],
            'y': [],
            't': []})

        nor_data: List = field(default_factory=lambda: {
            'x': [],
            'c': [],
            'y': [],
            't': []})

        rank: int = 2

class Trees():

    def __init__(self,
                 batch_size,
                 train_days,
                 test_days,
                 val_days,
                 seq_days,
                 seq_len,
                 seq_agg,
                 test_on,
                 val_on,
                 masking,
                 shuffle,
                 preproc,
                 seed,
                 name=None,
                 samples_pd=None,
                 photon=None,
                 network=None,
                 network_config=None):

        self.is_built = False

        if network is None:
            self.network = Networks(photon=photon, **network_config)
        else:
            self.network = network

        if name is None:
            self.name = 'Photon Tree'
        else:
            self.name = name

        self.val_on = val_on
        self.test_on = test_on

        self.n_branches = 0
        self.n_chains = 0

        self.branches = []
        self.chains = []

        self.tree_idx, self.data = self.network.add_tree(self)

        self.data.seq_days = seq_days
        self.data.seq_len = seq_len
        self.data.seq_agg = seq_agg
        self.data.seq_depth = seq_len
        self.data.res = self.network.data_res

        self.data.train_days = train_days
        self.data.test_days = test_days
        self.data.val_days = val_days

        self.data.masking = masking

        self.data.shuffle = shuffle
        self.data.preproc = preproc
        self.data.seed = seed

        self.data.batch_size = batch_size

        self.data.feature_cols = ['x_cols', 'c_cols']

        if samples_pd is None:

            self.data.samples_pd = int(23400 / self.data.res)
        else:
            self.data.samples_pd = samples_pd

    def load_data(self):

        self.data.preproc_trans = {
            'train': {'x_cols': None,
                      'c_cols': None,
                      'y_cols': None,
                      't_cols': None,
                      'pr_cols': None,
                      'vol_cols': None,
                      'atr_cols': None,
                      'roc_cols': None,
                      'zsc_cols': None},
            'test': {'y_cols': None,
                     't_cols': None},
            'val': {'y_cols': None,
                    't_cols': None}}

        self.data.store = {
            'train': {
                'close_bars': None,
                'model_bars': None,
                'x_bars': None,
                'y_bars': None,
                'c_bars': None,
                't_bars': None,
                'data_ds': None,
                'batch_data': None,
                'input_shp': None,
                'n_batches': None,
                'distributed_datasets': [],
                'config': {
                    'batch_size': self.data.batch_size,
                    'n_days': self.data.train_days,
                    'n_batches': 0,
                    'n_steps': 0,
                    'n_samples': 0,
                    'n_calls': 0,
                    'masks': {'blocks': {'all_cols': [],
                                         'x_cols': [],
                                         'y_cols': [],
                                         't_cols': []}}}},
            'test': {
                'close_bars': None,
                'model_bars': None,
                'data_ds': None,
                'batch_data': None,
                'input_shp': None,
                'n_batches': None,
                'config': {
                    'batch_size': self.data.batch_size,
                    'n_days': self.data.test_days,
                    'n_batches': 0,
                    'n_steps': 0,
                    'n_samples': 0,
                    'n_calls': 0,
                    'masks': {'blocks': {'all_cols': [],
                                         'x_cols': [],
                                         'y_cols': [],
                                         't_cols': []}}}},
            'val': {
                'data_bars': None,
                'close_bars': None,
                'model_bars': None,
                'data_ds': None,
                'batch_data': None,
                'input_shp': None,
                'n_batches': None,
                'config': {
                    'batch_size': self.data.batch_size,
                    'n_days': self.data.val_days,
                    'n_batches': 0,
                    'n_steps': 0,
                    'n_samples': 0,
                    'n_calls': 0,
                    'masks': {'blocks': {'all_cols': [],
                                         'x_cols': [],
                                         'y_cols': [],
                                         't_cols': []}}}}}

        # -- setup data -- #
        self.setup_data()

        self.types_on = ['train']

        if self.val_on:
            self.types_on.append('val')

        if self.test_on:
            self.types_on.append('test')

        # -- split bars -- #
        self.split_bars(self.types_on)

        # -- loop types -- #
        for data_type in self.types_on:
            self.setup_stores(data_type)
            self.pre_build_datasets(data_type)

        return

    def setup_data(self):

        self.data.dtype = self.network.dtype

        if self.data.seq_agg > 0:
            self.data.seq_depth = int(self.data.seq_len / self.data.seq_agg)

        if 'c_cols' in self.data.feature_cols:
            self.data.input_shape = (self.data.seq_depth, self.data.n_x_cols + self.data.n_c_cols)

        if 'c_cols' not in self.data.feature_cols:
            self.data.input_shape = (self.data.seq_depth, self.data.n_x_cols)

    def split_bars(self, _types):

        train_days = self.data.store['train']['config']['n_days']
        test_days = self.data.store['test']['config']['n_days']
        val_days = self.data.store['val']['config']['n_days']

        load_days = train_days + test_days + val_days

        self.data.close_bars = self.data.full_bars[self.data.full_bars['is_close'] == True][self.data.close_cols].copy()

        max_days = self.data.full_bars['day_idx'].max()

        full_base = max_days - load_days - 2

        train_st_day = full_base
        train_ed_day = train_st_day + train_days

        test_st_day = train_ed_day + 1

        if 'test' in _types:
            test_ed_day = test_st_day + test_days

        if 'test' not in _types:
            test_ed_day = test_st_day + 1

        if 'val' in _types:
            val_st_day = test_ed_day + 1
            val_ed_day = val_st_day + val_days

        # --- reduce full bars by number of days --- #

        self.data.store['train']['full_bars'] = \
            self.data.full_bars[(self.data.full_bars['day_idx'] >= (train_st_day - self.data.seq_days)) &
                                (self.data.full_bars['day_idx'] < train_ed_day)].copy()

        if 'test' in _types:
            self.data.store['test']['full_bars'] = \
                self.data.full_bars[(self.data.full_bars['day_idx'] >= (test_st_day - self.data.seq_days)) &
                                    (self.data.full_bars['day_idx'] < test_ed_day)].copy()

        if 'val' in _types:
            self.data.store['val']['full_bars'] = \
                self.data.full_bars[(self.data.full_bars['day_idx'] >= (val_st_day - self.data.seq_days)) &
                                    (self.data.full_bars['day_idx'] < val_ed_day)].copy()

        # --- reduce close bars by number of days --- #

        self.data.store['train']['close_bars'] = \
            self.data.close_bars[(self.data.close_bars['day_idx'] >= train_st_day) &
                                 (self.data.close_bars['day_idx'] < train_ed_day)].copy()

        if 'test' in _types:
            self.data.store['test']['close_bars'] = \
                self.data.close_bars[(self.data.close_bars['day_idx'] >= test_st_day) &
                                     (self.data.close_bars['day_idx'] < test_ed_day)].copy()

        if 'val' in _types:
            self.data.store['val']['close_bars'] = \
                self.data.close_bars[(self.data.close_bars['day_idx'] >= val_st_day) &
                                     (self.data.close_bars['day_idx'] < val_ed_day)].copy()

    def setup_stores(self, data_type):

        self.data.store[data_type]['config']['n_samples'] = \
            self.data.samples_pd * self.data.store[data_type]['config']['n_days']

        close_bars = self.data.store[data_type]['close_bars']

        full_bars = self.data.store[data_type]['full_bars']

        model_bars = []

        n_bars = close_bars.shape[0]

        # -- aggregate full bars -- #
        if self.data.seq_agg > 1:
            full_bars = self.agg_bars(full_bars)

        # -- normalise full bars -- #
        full_bars = self.normalize_bars(full_bars)

        # === append seq bars to close bars to generate model bars === #
        for _idx in range(n_bars):

            _bar = close_bars.iloc[_idx]

            _st_idx = _bar['bar_idx'] - self.data.seq_len
            _ed_idx = _bar['bar_idx']

            seq_bars = full_bars[(full_bars['bar_idx'] > _st_idx) &
                                 (full_bars['bar_idx'] <= _ed_idx)]

            model_bars.append(seq_bars.to_numpy())

            # self.data.setup_log['close_loop'].append(
            #     {'specs': {'bar': _bar, 'st_idx':_st_idx ,'ed_idx':_ed_idx},
            #      'seq_bars': seq_bars,
            #      'model_bars': model_bars})

        # -- concat model bars -- #
        model_bars = np.concatenate(model_bars, axis=0)

        # --- reshape seq bars ---  #
        n_samples = self.data.store[data_type]['config']['n_samples']

        seq_depth = int(model_bars.shape[0] / n_samples)

        _new_shp = (n_samples, seq_depth, model_bars.shape[-1])

        # -- reshape model bars -- #
        model_bars = np.reshape(model_bars, _new_shp)

        self.data.store[data_type]['full_bars'] = full_bars
        self.data.store[data_type]['model_bars'] = model_bars

        return

    def agg_bars(self, data_bars):

        n_bins = int(data_bars.shape[0] / self.data.seq_agg)

        data_bars = data_bars.groupby(pd.cut(data_bars.index, bins=n_bins)).agg(self.data.agg_data).dropna()

        return data_bars.reset_index(drop=True)

    def normalize_bars(self, data_bars):

        x_norm = self.data.preproc['normalize']['x_cols']
        c_norm = self.data.preproc['normalize']['c_cols']

        if x_norm is not None:

            # -- x groups off -- #
            if not self.network.x_groups_on:
                x_cols = self.data.nor_data['x']

                self.data.preproc_trans['train']['x_cols'] = getattr(preprocessing, x_norm['cls'])(**x_norm['params'])
                self.data.preproc_trans['train']['x_cols'].fit(data_bars[x_cols])
                data_bars[x_cols] = self.data.preproc_trans['train']['x_cols'].transform(data_bars[x_cols])

            # -- x groups on -- #
            if self.network.x_groups_on:
                # -- pr cols -- #
                pr_cols = self.data.x_groups['pr_group']

                self.data.preproc_trans['train']['pr_cols'] = getattr(preprocessing, x_norm['cls'])(**x_norm['params'])
                self.data.preproc_trans['train']['pr_cols'].fit(data_bars[pr_cols])
                data_bars[pr_cols] = self.data.preproc_trans['train']['pr_cols'].transform(data_bars[pr_cols])

                # -- vol cols -- #
                vol_cols = self.data.x_groups['vol_group']

                self.data.preproc_trans['train']['vol_cols'] = getattr(preprocessing, x_norm['cls'])(**x_norm['params'])
                self.data.preproc_trans['train']['vol_cols'].fit(data_bars[vol_cols])
                data_bars[vol_cols] = self.data.preproc_trans['train']['vol_cols'].transform(data_bars[vol_cols])

                # -- atr cols -- #
                atr_cols = self.data.x_groups['atr_group']

                self.data.preproc_trans['train']['atr_cols'] = getattr(preprocessing, x_norm['cls'])(**x_norm['params'])
                self.data.preproc_trans['train']['atr_cols'].fit(data_bars[atr_cols])
                data_bars[atr_cols] = self.data.preproc_trans['train']['atr_cols'].transform(data_bars[atr_cols])

                # -- roc cols -- #
                roc_cols = self.data.x_groups['roc_group']

                self.data.preproc_trans['train']['roc_cols'] = getattr(preprocessing, x_norm['cls'])(**x_norm['params'])
                self.data.preproc_trans['train']['roc_cols'].fit(data_bars[roc_cols])
                data_bars[roc_cols] = self.data.preproc_trans['train']['roc_cols'].transform(data_bars[roc_cols])

                # -- zsc cols -- #
                zsc_cols = self.data.x_groups['zsc_group']

                self.data.preproc_trans['train']['zsc_cols'] = getattr(preprocessing, x_norm['cls'])(**x_norm['params'])
                self.data.preproc_trans['train']['zsc_cols'].fit(data_bars[zsc_cols])
                data_bars[zsc_cols] = self.data.preproc_trans['train']['zsc_cols'].transform(data_bars[zsc_cols])

        # --- c cols --- #
        if c_norm is not None and self.data.n_c_cols > 0:
            c_cols = self.data.nor_data['c']

            self.data.preproc_trans['train']['c_cols'] = getattr(preprocessing, c_norm['cls'])(**c_norm['params'])
            self.data.preproc_trans['train']['c_cols'].fit(data_bars[c_cols])
            data_bars[c_cols] = self.data.preproc_trans['train']['c_cols'].transform(data_bars[c_cols])

        return data_bars

    def pre_build_datasets(self, data_type):

        batch_size = self.data.store[data_type]['config']['batch_size']

        n_batches = int(self.data.store[data_type]['model_bars'].shape[0] / batch_size)

        self.data.store[data_type]['n_batches'] = n_batches

        # -- splts -- #

        self.data.store[data_type]['x_bars'] = self.data.store[data_type]['model_bars'][..., :32]
        self.data.store[data_type]['c_bars'] = self.data.store[data_type]['model_bars'][..., 32:39]
        self.data.store[data_type]['y_bars'] = self.data.store[data_type]['model_bars'][..., 39:51]
        self.data.store[data_type]['t_bars'] = self.data.store[data_type]['model_bars'][..., 51:]

        if 'c_cols' in self.data.feature_cols:
            self.data.store[data_type]['x_bars'] = \
                np.concatenate([self.data.store[data_type]['x_bars'],
                                self.data.store[data_type]['c_bars']], axis=-1)

        x_ds = tf.data.Dataset.from_tensor_slices(tf.convert_to_tensor(self.data.store[data_type]['x_bars'],
                                                                       dtype=self.data.dtype))

        y_ds = tf.data.Dataset.from_tensor_slices(tf.convert_to_tensor(self.data.store[data_type]['y_bars'],
                                                                       dtype=self.data.dtype))

        t_ds = tf.data.Dataset.from_tensor_slices(tf.convert_to_tensor(self.data.store[data_type]['t_bars'],
                                                                       dtype=self.data.dtype))

        self.data.store[data_type]['data_ds'] = tf.data.Dataset.zip((x_ds, y_ds, t_ds)).batch(batch_size)

        # self.data.store[data_type]['data_ds'] = self.data.store[data_type]['data_ds'].prefetch(10)

        return

class Branches():

    def __init__(self,
                 trees,
                 n_epochs,
                 n_chains,
                 model_config,
                 data_config,
                 build_config,
                 opt_config,
                 loss_config,
                 metrics_config,
                 run_config,
                 save_config,
                 name=None,
                 **kwargs):

        self.is_built = False

        self.trees = trees

        self.network = self.trees[0].network
        self.photon = self.network.photon

        if name is None:
            self.name = 'Photon Branch'
        else:
            self.name = name

        self.n_epochs = n_epochs

        self.n_chains = n_chains
        self.chains = []

        self.msgs_on = False

        self.run = None

        # -- turn on branch msgs -- #
        for rc in run_config:
            if rc['msgs_on']:
                self.msgs_on = True

        # -- configs -- #
        self.configs = self.Configs(model_config=model_config,
                                    data_config=data_config,
                                    build_config=build_config,
                                    opt_config=opt_config,
                                    loss_config=loss_config,
                                    metrics_config=metrics_config,
                                    run_config=run_config,
                                    save_config=save_config)

        # --- add branch to network --- #
        self.branch_idx = self.network.add_branch(self)

        # --- loop chains to init/build --- #
        for chain_idx in range(self.n_chains):

            # -- init chains -- #
            chain = Chains(self, chain_idx)

            if not chain.is_built:
                chain.build_chain()

            self.chains.insert(chain_idx, chain)

    @dataclass()
    class Configs:

        model_config: List
        data_config: List
        build_config: List
        opt_config: List
        loss_config: List
        metrics_config: List
        run_config: List
        save_config: List

        def by_chain_idx(self, type:str, chain_idx:int) -> Dict:

            obj = getattr(self, type+'_config')

            if type == 'metrics':
                obj = [obj]

            if len(obj) <= chain_idx:
                obj_out = obj[-1]
            else:
                obj_out = obj[chain_idx]

            if len(obj) <= chain_idx:
                obj_out = obj[-1]
            else:
                obj_out = obj[chain_idx]

            return obj_out

class Chains():

    def __init__(self, branch, chain_idx):

        self.run_device_types = 'model_gpus'

        self.is_built = False

        self.branch = branch
        self.chain_idx = chain_idx

        self.network = self.branch.network
        self.network.chains.append(self)

        self.photon = self.branch.photon

        self.trees = []

        for tree in self.branch.trees:

            tree.chains.append(self)
            self.trees.append(tree)

        self.name = self.branch.name + '_chain_' + str(chain_idx)

        self.models = []

        self.datasets = []

        self.idx_gen = []

        self.input_data = None

        self.model_gpus = []

        self.model_subs = []

    def build_chain(self):

        self.model_config = self.branch.configs.by_chain_idx('model', self.chain_idx)
        self.data_config = self.branch.configs.by_chain_idx('data', self.chain_idx)
        self.build_config = self.branch.configs.by_chain_idx('build', self.chain_idx)

        self.opt_config = self.branch.configs.by_chain_idx('opt', self.chain_idx)
        self.loss_config = self.branch.configs.by_chain_idx('loss', self.chain_idx)
        self.metrics_config = self.branch.configs.by_chain_idx('metrics', self.chain_idx)
        self.save_config = self.branch.configs.by_chain_idx('save', self.chain_idx)

        self.n_models = self.model_config['n_models']
        self.n_outputs = self.model_config['n_models']

        # -- setup GPUS -- #
        self.map_gpus()

        # -- setup data slices -- #
        self.setup_slices()

        # -- loop trees -- #
        for tree_idx, tree in enumerate(self.trees):

            # -- load data -- #
            tree.load_data()

            # --- setup datasets --- #
            train_ds = tree.data.store['train']['data_ds']
            val_ds = None
            test_ds = None

            if tree.data.store['val']['data_ds']:
                val_ds = tree.data.store['val']['data_ds']

            if tree.data.store['test']['data_ds']:
                test_ds = tree.data.store['test']['data_ds']

            self.datasets.insert(tree_idx, {'train': train_ds,
                                            'val': val_ds,
                                            'test': test_ds})

        # -- build input data placeholder -- #
        if self.input_data is None:

            self.input_data = tf.keras.Input(shape=self.trees[0].data.input_shape,
                                             batch_size=self.trees[0].data.batch_size,
                                             dtype=self.network.float_x,
                                             name=self.name + '_input_data')

        # --- setup gauge/models --- #
        for model_idx in range(self.n_models):

            # --- init gauge --- #
            gauge = Gauge(chain=self, model_idx=model_idx)

            # # --- build gauge --- #
            # if gauge.strat_on:
            #     with gauge.strat.scope():
            #         gauge.build_gauge(self)
            #
            # if not gauge.strat_on:
            #     gauge.build_gauge(self)

            # -- insert into chain models -- #
            self.models.insert(model_idx, gauge)

        self.is_built = True

        return

    def map_gpus(self):

        if self.photon.n_v_gpus == 0 and self.photon.n_gpus > 1:

            grp_idx = 0

            for model_idx in range(self.n_models):

                dis = model_idx % self.photon.n_gpus
                grp_dis = grp_idx % 2

                gpu_idx = 0

                if grp_dis == 0:
                    gpu_idx = dis


                # print(f'{dis} {grp_dis} {grp_idx} {gpu_idx}')

                gpu_data = {'gpu_idx': gpu_idx,
                            'model_idx': model_idx,
                            'v_run_device': '/device:GPU:' + str(gpu_idx)}

                self.model_gpus.insert(model_idx, gpu_data)

                grp_idx += 3

        if self.photon.n_v_gpus > 0:

            grp_idx = 0
            grp_mod = 0

            for model_idx in range(self.n_models):

                if model_idx >= self.photon.n_v_gpus:

                    grp_idx += 1

                    grp_mod = math.ceil(grp_idx/self.photon.n_v_gpus)

                gpu_idx = model_idx - (self.photon.n_v_gpus * grp_mod)

                gpu_data = {'gpu_idx': gpu_idx,
                            'model_idx': model_idx,
                            'v_run_device': self.photon.v_gpus[gpu_idx]['v_run_device']}

                self.model_gpus.insert(model_idx, gpu_data)

        return

    def setup_slices(self):

        targets_config = self.data_config['targets']
        split_on = targets_config['split_on']

        self.data_config['targets']['true_slice'] = np.s_[...,:split_on]
        self.data_config['targets']['tracking_slice'] = np.s_[..., split_on:]

class Gauge():

    def __init__(self, chain, model_idx):

        self.is_built = False
        self.is_compiled = False

        self.is_model_built = False

        self.chain = chain
        self.model_idx = model_idx

        self.chain_idx = self.chain.chain_idx
        self.branch = self.chain.branch
        self.network = self.branch.network
        self.dtype = self.network.dtype

        self.name = chain.name + '_model_' + str(model_idx)

        self.input_shape = ()

        self.layers = {}
        self.parent_layers = {}
        self.child_layers = {}

        self.src = None

        self.model_args = None
        self.model_inputs = None

        self.opt_fn = None
        self.loss_fn = None
        self.metrics_fn = None

        self.conn_chain_idx = None
        self.conn_n_models = None

        self.chkp_manager = None
        self.chkp_dir = None

        self.run_chain = None
        self.run_model = None
        self.run_data = None
        self.is_live = False

        self.strat_on = False
        self.dist_on = False

        self.runs = []

        self.batch_data = []
        self.val_batch_data = []

        self.datasets = {'train': None,
                         'val': None,
                         'test': None}

        # -- setup strats -- #
        self.setup_strats()

    def setup_strats(self):

        self.strat_type = self.chain.build_config['strat_type']
        self.dist_type = self.chain.build_config['dist_type']

        # self.run_device = '/GPU:0'
        self.strat = None

        if self.chain.photon.n_gpus > 1 or self.chain.photon.n_v_gpus > 0:

            # if self.strat_type is None:
            # self.run_device = self.chain.model_gpus[self.model_idx]['v_run_device']

            if self.strat_type is not None:

                self.strat_on = True

                if self.strat_type == 'Mirrored':
                    self.strat = tf.distribute.MirroredStrategy(self.chain.photon.gpus)
                    self.dist_on = True

        # if self.strat_type == 'One':
        #
        #     self.strat_on = True
        #     self.strat = tf.distribute.OneDeviceStrategy(self.run_device)

    def build_gauge(self, run, chain, model):

        self.tree = self.chain.trees[0]

        self.model_args = self.chain.model_config['args']

        # -- init model -- #
        self.src = self.chain.model_config['model'](**{'gauge': self})

        # -- opt_fn -- #
        self.opt_fn = self.chain.opt_config['fn']

        # -- setup loss fn -- #
        self.setup_loss_fn()

        self.is_built = True

        self.runs.insert(run.run_idx, run)

        self.run_chain = chain
        self.run_model = model

        return

    def setup_loss_fn(self):

        _loss_reduc = self.chain.loss_config['args']['reduction']
        _loss_from_logits = self.chain.loss_config['args']['from_logits']

        if _loss_reduc == 'NONE':
            self.loss_fn = self.chain.loss_config['fn'](from_logits=_loss_from_logits,
                                                        reduction=tf.keras.losses.Reduction.NONE)

        if _loss_reduc == 'SUM':
            self.loss_fn = self.chain.loss_config['fn'](from_logits=_loss_from_logits,
                                                        reduction=tf.keras.losses.Reduction.SUM)

        if _loss_reduc == 'AUTO':
            self.loss_fn = self.chain.loss_config['fn'](from_logits=_loss_from_logits,
                                                        reduction=tf.keras.losses.Reduction.AUTO)

        if _loss_reduc == 'SUM_OVER_BATCH_SIZE':
            self.loss_fn = self.chain.loss_config['fn'](from_logits=_loss_from_logits,
                                                         reduction=tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE)

    def compile_gauge(self, tree_idx):

        self.opt_fn = self.opt_fn(gauge=self,
                                  tree=self.chain.trees[tree_idx],
                                  config=self.chain.opt_config['args'],
                                  n_epochs=self.branch.n_epochs)

        self.metrics_fns = []

        for config in self.chain.metrics_config:
            self.metrics_fns.append(config['fn'](config=config['args']))

        # -- compile model -- #
        self.src.compile(optimizer=self.opt_fn)

        self.is_compiled = True

    def setup_cp(self, step_idx, load_cp):

        if not self.is_compiled:
            self.compile_gauge()

        self.chkp_nm = self.chain.name + '_model_' + str(self.model_idx)

        self.chkp_dir = self.get_chkp_dir()

        self.chkp = tf.train.Checkpoint(model=self.src,
                                        step_idx=step_idx)

        self.chkp_manager = tf.train.CheckpointManager(checkpoint=self.chkp,
                                                       directory=self.chkp_dir,
                                                       max_to_keep=5,
                                                       step_counter=step_idx,
                                                       checkpoint_name=self.chkp_nm,
                                                       checkpoint_interval=1)

        if self.chkp_manager.latest_checkpoint and load_cp:

            chkp_status = None
            chkp_status = self.chkp.restore(self.chkp_manager.latest_checkpoint)
            chkp_status.assert_existing_objects_matched()

            print(f'model restored from {self.chkp_manager.latest_checkpoint}')

    def get_chkp_dir(self):

        branch_nm = 'branch_' + str(self.branch.branch_idx)

        chkp_dir = self.network.photon.store['chkps'] + '/' + self.network.photon.photon_nm.lower() + '/' + branch_nm + '/' + self.chain.name.lower() + '/model_' + str(self.model_idx)

        if self.network.photon_load_id == 0 and os.path.exists(chkp_dir):
            os.remove(chkp_dir)

        return chkp_dir

    def setup_run(self, run, chain, model):

        self.runs.insert(run.run_idx, run)

        self.run_chain = chain
        self.run_model = model

    def pre_build_model(self):

        if not self.is_compiled:
            self.compile_gauge()

        self.src.pre_build(input_data=self.chain.input_data)

    def run_steps(self):

        # self.run_train_steps(inputs)

        return

    def run_train_steps_(self, inputs, targets, batch_idx):

        # tf.print('-- ',self.name, self.run_train_steps.experimental_get_tracing_count())
                 # , batch_idx, inputs.shape, targets.shape, '\n')

        batch_size = self.chain.branch.trees[0].data.batch_size

        # --- load tape --- #
        with tf.GradientTape() as tape:

            # -- run model -- #
            step_data = self.src(inputs=inputs, training=True, batch_idx=batch_idx)

            # -- slice y_true -- #
            step_data['y_true'] = targets[self.chain.data_config['targets']['true_slice']]

            # -- slice y_tracking
            step_data['y_tracking'] = targets[self.chain.data_config['targets']['tracking_slice']]

            # -- step loss -- #
            step_data['step_loss'] = self.loss_fn(step_data['y_true'], step_data['y_hat'])
            step_data['step_loss'] = tf.nn.compute_average_loss(step_data['step_loss'], global_batch_size=batch_size)

            # -- model loss -- #
            step_data['model_loss'] = sum(self.src.losses)

            # -- full loss -- #
            step_data['full_loss'] = tf.add(step_data['step_loss'], step_data['model_loss'])

            # -- run grads -- #
            step_data['step_grads'] = tape.gradient(step_data['step_loss'], self.src.trainable_variables)

            # -- apply grads -- #
            self.src.optimizer.apply_gradients(zip(step_data['step_grads'], self.src.trainable_variables))

            # -- save lr -- #
            step_data['learning_rates'] = self.src.optimizer.lr_sch.cur_lr

        return step_data

    def run_preds_steps_(self):

        tf.print('-- VAL --- ',self.name, self.run_preds_steps.python_function)
                 # batch_idx, inputs.shape, targets.shape, '\n')

        # batch_size = self.chain.branch.trees[0].data.batch_size

        # -- run model -- #
        # step_data = self.src(inputs=inputs, training=False, batch_idx=batch_idx)

        some_data = ['sss']

        # # -- slice y_true -- #
        # step_data['y_true'] = targets[model.chain.data_config['targets']['true_slice']]
        #
        # # -- slice y_tracking
        # step_data['y_tracking'] = targets[model.chain.data_config['targets']['tracking_slice']]
        #
        # # -- step loss -- #
        # step_data['step_loss'] = model.gauge.loss_fn(step_data['y_true'], step_data['y_hat'])
        # step_data['step_loss'] = tf.nn.compute_average_loss(step_data['step_loss'], global_batch_size=batch_size)
        #
        # # -- model loss -- #
        # step_data['model_loss'] = sum(model.losses)
        #
        # # -- full loss -- #
        # step_data['full_loss'] = tf.add(step_data['step_loss'], step_data['model_loss'])
        #
        # step_data['step_grads'] = None
        # step_data['learning_rates'] = None

        return some_data

class Theta:

    def __init__(self, run):

        self.run = run

        self.params = {'model_pre': [], 'model_post': [], 'opt': [], 'grads': []}

        self.logs = []

    def save_params(self, param_type, grads=None):

        if self.run.cur_chain.logs_on['theta']:

            if param_type == 'model_pre' or param_type == 'model_post':

                for idx, p in enumerate(self.run.model.trainable_variables):

                    p_name = p.name

                    p_data = {'idx': idx,
                              'name': p_name,
                              'shape': p.shape,
                              'avg': tf.math.reduce_mean(p).numpy(),
                              'value': p.numpy()}

                    self.params[param_type].append(p_data)

                    if grads is not None:

                        if grads[idx] is not None:
                            grads_data = {'idx': idx,
                                          'name': p_name,
                                          'shape': grads[idx].shape,
                                          'avg': tf.math.reduce_mean(grads[idx]).numpy(),
                                          'value': grads[idx].numpy()}

                            self.params['grads'].append(grads_data)

            if param_type == 'opt':

                for idx, opt in enumerate(self.run.model.optimizer.get_weights()):

                    if idx > 0:
                        opt_data = {'idx': idx,
                                    'shape': opt.shape,
                                    'avg': tf.math.reduce_mean(opt).numpy(),
                                    'value': opt}

                        self.params['opt'].append(opt_data)

    def log_params(self):

        if self.run.cur_chain.logs_on['theta']:

            epoch_idx = self.run.epoch_idx

            if len(self.logs) <= epoch_idx:
                self.logs.append([])

            self.logs[epoch_idx].append(self.params.copy())

            self.params = {'model_pre': [], 'model_post': [], 'opt': [], 'grads': []}


import os
import tensorflow as tf
import numpy as np
import pandas as pd

def dest_append(obj, dest, key=None, epoch_idx=None, batch_idx=None):

    if key is not None:
        dest = getattr(dest, key)

    if epoch_idx is not None:

        if len(dest) <= epoch_idx:
            dest.append([])

        if batch_idx is not None:

            dest[epoch_idx].insert(batch_idx, obj)

            # if len(dest[epoch_idx]) <= batch_idx:
            #     dest[epoch_idx].append([])

        # dest[epoch_idx][batch_idx] = obj

        return dest

def np_exp(obj):
    if 'numpy' in dir(obj):
        return obj.numpy()
    else:
        return obj

def obj_exp(obj, strat=None, to_numpy=True, reduce=None, axis=None, list_out=False):

    obj_out = obj

    if isinstance(obj_out, list) and not list_out:

            obj_out = list_exp(obj=obj_out, strat=strat, to_numpy=to_numpy, reduce=reduce, axis=axis)

    if strat is not None:
        if isinstance(obj_out, tf.distribute.DistributedValues):

            if reduce is None:
                obj_out = tf.concat(strat.experimental_local_results(obj_out), axis=axis)

            if reduce is not None:
                if str.upper(reduce) == 'SUM':
                    obj_out = strat.reduce(tf.distribute.ReduceOp.SUM, obj_out, axis=axis)

                if str.upper(reduce) == 'MEAN':
                    obj_out = strat.reduce(tf.distribute.ReduceOp.MEAN, obj_out, axis=axis)

    if to_numpy:
        obj_out = np_exp(obj_out)

    return obj_out

def list_exp(obj, strat=None, to_numpy=True, reduce=None, axis=None):

    obj_out = []

    if isinstance(obj, list):

        for _obj in obj:
            obj_out.append(obj_exp(obj=obj_out, strat=strat, to_numpy=to_numpy, reduce=reduce, axis=axis))

    return obj_out

def config_block_mask(off_blocks=None):

    ''' mask on means block should be utilized, mask off means it will be skipped '''

    block_times = [
        [1, 093001.00, 100000.00],
        [2, 100001.00, 103000.00],
        [3, 103001.00, 110000.00],
        [4, 110001.00, 113000.00],
        [5, 113001.00, 120000.00],
        [6, 120001.00, 123000.00],
        [7, 123001.00, 130000.00],
        [8, 130001.00, 133000.00],
        [9, 133001.00, 140000.00],
        [10, 140001.00, 143000.00],
        [11, 143001.00, 150000.00],
        [12, 150001.00, 153000.00],
        [13, 153001.00, 160000.00]]

    _mask = np.ones(len(block_times), dtype=bool)

    if off_blocks is not None:

        off_blocks = np.array(off_blocks) - 1

        _mask[off_blocks] = False

    return _mask


import tensorflow as tf
import numpy as np

# ::: Metrics Class ::: #
class Metrics():

    def __init__(self):

        self.CatAcc = CatAcc
        self.Precision = Precision

class CatAcc():

    def __init__(self, config):

        self.config = config

        self.main_acc = tf.keras.metrics.CategoricalAccuracy()
        self.val_acc = tf.keras.metrics.CategoricalAccuracy()

    def __call__(self, y_true, y_hat, run_type):

        if run_type != 'val':

            self.main_acc.update_state(y_true, y_hat)

            return self.main_acc.result().numpy()

        if run_type == 'val':

            self.val_acc.update_state(y_true, y_hat)

            return self.val_acc.result().numpy()

class Precision():

    def __init__(self, config):

        self.config = config

        self.main_acc = tf.keras.metrics.Precision()
        self.val_acc = tf.keras.metrics.Precision()

    def __call__(self, y_true, y_hat, run_type):

        if run_type != 'val':

            self.main_acc.update_state(y_true, tf.nn.softmax(y_hat))

            return self.main_acc.result().numpy()

        if run_type == 'val':

            self.val_acc.update_state(y_true, tf.nn.softmax(y_hat))

            return self.val_acc.result().numpy()

